from sqlalchemy.orm import aliased
from sqlalchemy.exc import StatementError

from zou.app.utils import (
    cache,
    events,
    fields,
    query as query_utils,
)

from zou.app.models.entity import (
    Entity,
    EntityLink,
    EntityVersion,
    EntityConceptLink,
)
from zou.app.models.project import Project
from zou.app.models.subscription import Subscription
from zou.app.models.task import Task, TaskPersonLink

from zou.app.services import (
    deletion_service,
    entities_service,
    notifications_service,
    user_service,
)
from zou.app.services.exception import (
    EditNotFoundException,
    WrongIdFormatException,
)


def clear_edit_cache(edit_id):
    cache.cache.delete_memoized(get_edit, edit_id)
    cache.cache.delete_memoized(get_edit, edit_id, True)
    cache.cache.delete_memoized(get_full_edit, edit_id)


@cache.memoize_function(1200)
def get_edit_type():
    return entities_service.get_temporal_entity_type_by_name("Edit")


def get_edits(criterions={}):
    """
    Get all edits for given criterions.
    """
    edit_type = get_edit_type()
    criterions["entity_type_id"] = edit_type["id"]
    is_only_assignation = "assigned_to" in criterions
    if is_only_assignation:
        del criterions["assigned_to"]

    query = Entity.query
    query = query_utils.apply_criterions_to_db_query(Entity, query, criterions)
    query = query.join(Project).add_columns(Project.name).order_by(Entity.name)

    if is_only_assignation:
        query = query.outerjoin(Task, Task.entity_id == Entity.id)
        query = query.filter(user_service.build_assignee_filter())

    try:
        data = query.all()
    except StatementError:  # Occurs when an id is not properly formatted
        raise WrongIdFormatException

    edits = []
    for edit_model, project_name in data:
        edit = edit_model.serialize(obj_type="Edit")
        edit["project_name"] = project_name
        edits.append(edit)

    return edits


def get_edits_and_tasks(criterions={}):
    """
    Get all edits for given criterions with related tasks for each edit.
    """
    edit_type = get_edit_type()
    edit_map = {}
    task_map = {}
    Episode = aliased(Entity, name="episode")
    subscription_map = notifications_service.get_subscriptions_for_user(
        criterions.get("project_id", None), get_edit_type()["id"]
    )

    query = (
        Entity.query.join(Project, Entity.project_id == Project.id)
        .outerjoin(Episode, Episode.id == Entity.parent_id)
        .outerjoin(Task, Task.entity_id == Entity.id)
        .outerjoin(TaskPersonLink)
        .add_columns(
            Episode.id,
            Episode.name,
            Task.id,
            Task.task_type_id,
            Task.task_status_id,
            Task.priority,
            Task.estimation,
            Task.duration,
            Task.retake_count,
            Task.real_start_date,
            Task.end_date,
            Task.start_date,
            Task.due_date,
            Task.last_comment_date,
            Task.last_preview_file_id,
            Task.nb_assets_ready,
            TaskPersonLink.person_id,
            Project.id,
            Project.name,
        )
        .filter(Entity.entity_type_id == edit_type["id"])
    )
    if "id" in criterions:
        query = query.filter(Entity.id == criterions["id"])

    if "project_id" in criterions:
        query = query.filter(Entity.project_id == criterions["project_id"])

    if "episode_id" in criterions:
        query = query.filter(Entity.parent_id == criterions["episode_id"])

    if "assigned_to" in criterions:
        query = query.filter(user_service.build_assignee_filter())
        del criterions["assigned_to"]

    query_result = query.all()

    if "vendor_departments" in criterions:
        not_allowed_descriptors_field_names = (
            entities_service.get_not_allowed_descriptors_fields_for_vendor(
                "Edit",
                criterions["vendor_departments"],
                set(edit[0].project_id for edit in query_result),
            )
        )

    for (
        edit,
        episode_id,
        episode_name,
        task_id,
        task_type_id,
        task_status_id,
        task_priority,
        task_estimation,
        task_duration,
        task_retake_count,
        task_real_start_date,
        task_end_date,
        task_start_date,
        task_due_date,
        task_last_comment_date,
        task_last_preview_file_id,
        task_nb_assets_ready,
        person_id,
        project_id,
        project_name,
    ) in query.all():
        edit_id = str(edit.id)

        if edit_id not in edit_map:
            data = fields.serialize_value(edit.data or {})
            if "vendor_departments" in criterions:
                data = (
                    entities_service.remove_not_allowed_fields_from_metadata(
                        not_allowed_descriptors_field_names[edit.project_id],
                        data,
                    )
                )

            edit_map[edit_id] = fields.serialize_dict(
                {
                    "canceled": edit.canceled,
                    "data": data,
                    "description": edit.description,
                    "entity_type_id": edit.entity_type_id,
                    "episode_id": episode_id,
                    "episode_name": episode_name or "",
                    "id": edit.id,
                    "name": edit.name,
                    "parent_id": edit.parent_id,
                    "preview_file_id": edit.preview_file_id or None,
                    "project_id": project_id,
                    "project_name": project_name,
                    "source_id": edit.source_id,
                    "nb_entities_out": edit.nb_entities_out,
                    "tasks": [],
                    "type": "Edit",
                }
            )

        if task_id is not None:
            task_id = str(task_id)
            if task_id not in task_map:
                task_dict = fields.serialize_dict(
                    {
                        "id": task_id,
                        "duration": task_duration,
                        "due_date": task_due_date,
                        "entity_id": edit_id,
                        "end_date": task_end_date,
                        "estimation": task_estimation,
                        "last_comment_date": task_last_comment_date,
                        "last_preview_file_id": task_last_preview_file_id,
                        "is_subscribed": subscription_map.get(task_id, False),
                        "nb_assets_ready": task_nb_assets_ready,
                        "priority": task_priority or 0,
                        "real_start_date": task_real_start_date,
                        "retake_count": task_retake_count,
                        "start_date": task_start_date,
                        "task_status_id": task_status_id,
                        "task_type_id": task_type_id,
                        "assignees": [],
                    }
                )
                task_map[task_id] = task_dict
                edit_dict = edit_map[edit_id]
                edit_dict["tasks"].append(task_dict)

            if person_id:
                task_map[task_id]["assignees"].append(str(person_id))

    return list(edit_map.values())


def get_edit_raw(edit_id):
    """
    Return given edit as an active record.
    """
    edit_type = get_edit_type()
    try:
        edit = Entity.get_by(entity_type_id=edit_type["id"], id=edit_id)
    except StatementError:
        raise EditNotFoundException

    if edit is None:
        raise EditNotFoundException

    return edit


@cache.memoize_function(120)
def get_edit(edit_id, relations=False):
    """
    Return given edit as a dictionary.
    """
    return get_edit_raw(edit_id).serialize(
        obj_type="Edit", relations=relations
    )


@cache.memoize_function(120)
def get_full_edit(edit_id):
    """
    Return given edit as a dictionary with extra data like project.
    """
    edits = get_edits_and_tasks({"id": edit_id})
    if len(edits) > 0:
        return edits[0]
    else:
        raise EditNotFoundException


def is_edit(entity):
    """
    Returns True if given entity has 'Edit' as entity type
    """
    edit_type = get_edit_type()
    return str(entity["entity_type_id"]) == edit_type["id"]


def get_edits_for_project(project_id, only_assigned=False):
    """
    Retrieve all edits related to given project.
    """
    return entities_service.get_entities_for_project(
        project_id, get_edit_type()["id"], "Edit", only_assigned=only_assigned
    )


def get_edits_for_episode(episode_id, relations=False):
    """
    Get all edits for given episode.
    """
    edit_type_id = get_edit_type()["id"]
    result = (
        Entity.query.filter(Entity.entity_type_id == edit_type_id).filter(
            Entity.parent_id == episode_id
        )
    ).all()
    return Entity.serialize_list(result, "Edit", relations=relations)


def remove_edit(edit_id, force=False):
    """
    Remove given edit from database. If it has tasks linked to it, it marks
    the edit as canceled. Deletion can be forced.
    """
    edit = get_edit_raw(edit_id)
    is_tasks_related = Task.query.filter_by(entity_id=edit_id).count() > 0

    if is_tasks_related and not force:
        edit.update({"canceled": True})
        clear_edit_cache(edit_id)
        events.emit(
            "edit:update",
            {"edit_id": edit_id},
            project_id=str(edit.project_id),
        )
    else:
        from zou.app.services import tasks_service

        tasks = Task.query.filter_by(entity_id=edit_id).all()
        for task in tasks:
            deletion_service.remove_task(task.id, force=True)
            tasks_service.clear_task_cache(str(task.id))

        EntityVersion.delete_all_by(entity_id=edit_id)
        Subscription.delete_all_by(entity_id=edit_id)
        EntityLink.delete_all_by(entity_in_id=edit_id)
        EntityLink.delete_all_by(entity_out_id=edit_id)
        EntityConceptLink.delete_all_by(entity_in_id=edit_id)
        EntityConceptLink.delete_all_by(entity_out_id=edit_id)

        edit.delete()
        clear_edit_cache(edit_id)
        events.emit(
            "edit:delete",
            {"edit_id": edit_id},
            project_id=str(edit.project_id),
        )

    deleted_edit = edit.serialize(obj_type="Edit")
    return deleted_edit


def create_edit(
    project_id, name, data={}, description="", parent_id=None, created_by=None
):
    """
    Create edit for given project and episode.
    """
    edit_type = get_edit_type()

    if parent_id is not None and len(parent_id) < 36:
        parent_id = None

    edit = Entity.get_by(
        entity_type_id=edit_type["id"],
        parent_id=parent_id,
        project_id=project_id,
        name=name,
    )
    if edit is None:
        edit = Entity.create(
            entity_type_id=edit_type["id"],
            project_id=project_id,
            parent_id=parent_id,
            name=name,
            data=data,
            description=description,
            created_by=created_by,
        )
    events.emit(
        "edit:new",
        {
            "edit_id": edit.id,
            "parent_id": parent_id,
        },
        project_id=project_id,
    )
    return edit.serialize(obj_type="Edit")


def update_edit(edit_id, data_dict):
    """
    Update fields of an edit with a given edit_id with data given data_dict.
    """
    edit = get_edit_raw(edit_id)
    edit.update(data_dict)
    clear_edit_cache(edit_id)
    events.emit(
        "edit:update", {"edit_id": edit_id}, project_id=str(edit.project_id)
    )
    return edit.serialize()


def get_edit_versions(edit_id):
    """
    Edit metadata changes are versioned. This function returns all versions
    of a given edit.
    """
    versions = (
        EntityVersion.query.filter_by(entity_id=edit_id)
        .order_by(EntityVersion.created_at.desc())
        .all()
    )
    return EntityVersion.serialize_list(versions, obj_type="EditVersion")



from zou.app.blueprints.export.csv.base import BaseCsvExport
from sqlalchemy.orm import aliased

from zou.app.models.task_status import TaskStatus
from zou.app.models.task_type import TaskType
from zou.app.models.task import Task
from zou.app.models.person import Person
from zou.app.models.project import Project
from zou.app.models.entity import Entity
from zou.app.models.entity_type import EntityType

from zou.app.services import projects_service


class TasksCsvExport(BaseCsvExport):
    def __init__(self):
        BaseCsvExport.__init__(self)

        self.file_name = "tasks_export"

    def build_headers(self):
        return [
            "Project",
            "Task Type",
            "Episode",
            "Sequence",
            "Entity Type",
            "Entity",
            "Assigner",
            "Assignees",
            "Duration",
            "Estimation",
            "Start date",
            "Due date",
            "WIP date",
            "Validation date",
            "Task Status",
        ]

    def build_query(self):
        Sequence = aliased(Entity, name="sequence")
        Episode = aliased(Entity, name="episode")
        open_status = projects_service.get_open_status()

        query = Task.query.order_by(Project.name, TaskType.name, Task.name)
        query = query.join(Project, Task.project_id == Project.id)
        query = query.join(TaskType, Task.task_type_id == TaskType.id)
        query = query.join(TaskStatus, Task.task_status_id == TaskStatus.id)
        query = query.join(Entity, Task.entity_id == Entity.id)
        query = query.join(EntityType)
        query = (
            query.outerjoin(Person, Task.assigner_id == Person.id)
            .outerjoin(Sequence, Sequence.id == Entity.parent_id)
            .outerjoin(Episode, Episode.id == Sequence.parent_id)
        )
        query = query.add_columns(
            Project.name,
            TaskType.name,
            TaskStatus.name,
            Episode.name,
            Sequence.name,
            EntityType.name,
            Entity.name,
            Person.first_name,
            Person.last_name,
        )
        query = query.filter(Project.project_status_id == open_status["id"])
        query = query.order_by(
            Project.name,
            TaskType.name,
            Episode.name,
            Sequence.name,
            EntityType.name,
            Entity.name,
        )

        return query

    def build_row(self, task_data):
        (
            task,
            project_name,
            task_type_name,
            task_status_name,
            episode_name,
            sequence_name,
            entity_type_name,
            entity_name,
            assigner_first_name,
            assigner_last_name,
        ) = task_data

        assigner_name = ""
        if assigner_first_name is not None and assigner_last_name is not None:
            assigner_name = assigner_first_name + " " + assigner_last_name
        elif assigner_first_name is not None:
            assigner_name = assigner_first_name
        elif assigner_last_name is not None:
            assigner_name = assigner_last_name

        persons = task.assignees_as_string()

        start_date = ""
        if task.start_date is not None:
            start_date = task.start_date.strftime("%Y-%m-%d")

        due_date = ""
        if task.due_date is not None:
            due_date = task.due_date.strftime("%Y-%m-%d")

        real_start_date = ""
        if task.real_start_date is not None:
            real_start_date = task.real_start_date.strftime("%Y-%m-%d")

        end_date = ""
        if task.end_date is not None:
            end_date = task.end_date.strftime("%Y-%m-%d")

        return [
            project_name,
            task_type_name,
            episode_name,
            sequence_name,
            entity_type_name,
            entity_name,
            assigner_name,
            persons,
            task.duration,
            task.estimation,
            start_date,
            due_date,
            real_start_date,
            end_date,
            task_status_name,
        ]



import datetime

from flask import abort, request, current_app
from flask_restful import Resource
from flask_jwt_extended import jwt_required

from babel.dates import format_datetime

from zou.app import config
from zou.app.mixin import ArgsMixin
from zou.app.services import (
    persons_service,
    tasks_service,
    time_spents_service,
    shots_service,
    user_service,
)
from zou.app.utils import (
    permissions,
    csv_utils,
    auth,
    emails,
    fields,
    date_helpers,
)
from zou.app.services.exception import (
    DepartmentNotFoundException,
    WrongDateFormatException,
    WrongParameterException,
    UnactiveUserException,
    TwoFactorAuthenticationNotEnabledException,
    PersonInProtectedAccounts,
)
from zou.app.services.auth_service import (
    disable_two_factor_authentication_for_person,
)


class DesktopLoginsResource(Resource, ArgsMixin):
    """
    Allow to create and retrieve desktop login logs. Desktop login logs can only
    be created by current user.
    """

    @jwt_required()
    def get(self, person_id):
        """
        Retrieve desktop login logs.
        ---
        tags:
        - Persons
        description: Desktop login logs can only be created by current user.
        parameters:
          - in: path
            name: person_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: Desktop login logs
        """
        current_user = persons_service.get_current_user()
        if (
            current_user["id"] != person_id
            and not permissions.has_manager_permissions()
        ):
            raise permissions.PermissionDenied

        persons_service.get_person(person_id)
        return persons_service.get_desktop_login_logs(person_id)

    @jwt_required()
    def post(self, person_id):
        """
        Create desktop login logs.
        ---
        tags:
        - Persons
        description: Add a new log entry for desktop logins.
        parameters:
          - in: path
            name: person_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: formData
            name: date
            required: True
            type: string
            format: date
            example: "2022-07-12"
        responses:
            201:
                description: Desktop login log entry created.
        """
        args = self.get_args([("date", date_helpers.get_utc_now_datetime())])

        current_user = persons_service.get_current_user()
        if (
            current_user["id"] != person_id
            and not permissions.has_admin_permissions()
        ):
            raise permissions.PermissionDenied

        desktop_login_log = persons_service.create_desktop_login_logs(
            person_id, args["date"]
        )

        return desktop_login_log, 201


class PresenceLogsResource(Resource):
    """
    Return a csv file containing the presence logs based on a daily basis.
    """

    @jwt_required()
    def get(self, month_date):
        """
        Return a csv file containing the presence logs based on a daily basis.
        ---
        tags:
        - Persons
        parameters:
          - in: path
            name: month_date
            required: True
            type: string
            format: date
            example: "2022-07"
        responses:
            200:
                description: CSV file containing the presence logs based on a daily basis
        """
        permissions.check_admin_permissions()
        date = datetime.datetime.strptime(month_date, "%Y-%m")
        presence_logs = persons_service.get_presence_logs(
            date.year, date.month
        )
        return csv_utils.build_csv_response(presence_logs)


class TimeSpentsResource(Resource, ArgsMixin):
    """
    Get all time spents for the given person.
    Optionnaly can accept date range parameters.
    """

    @jwt_required()
    def get(self, person_id):
        user_service.check_person_is_not_bot(person_id)
        permissions.check_admin_permissions()
        arguments = self.get_args(["start_date", "end_date"])
        start_date, end_date = arguments["start_date"], arguments["end_date"]
        if not start_date and not end_date:
            return time_spents_service.get_time_spents(person_id)

        if None in [start_date, end_date]:
            abort(
                400,
                "If querying for a range of dates, both a `start_date` and"
                " an `end_date` must be given.",
            )

        try:
            return time_spents_service.get_time_spents_range(
                person_id, start_date, end_date
            )
        except WrongDateFormatException:
            abort(
                400,
                f"Wrong date format for {start_date} and/or {end_date}",
            )


class DateTimeSpentsResource(Resource):
    """
    Get time spents for given person and date.
    """

    @jwt_required()
    def get(self, person_id, date):
        """
        Get time spents for given person and date.
        ---
        tags:
        - Persons
        parameters:
          - in: path
            name: person_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: date
            required: True
            type: string
            format: date
            example: "2022-07-12"
        responses:
            200:
                description: Time spents for given person and date
            404:
                description: Wrong date format
        """
        user_service.check_person_is_not_bot(person_id)
        department_ids = None
        project_ids = None
        if not permissions.has_admin_permissions():
            if persons_service.get_current_user()["id"] != person_id:
                if (
                    permissions.has_manager_permissions()
                    or permissions.has_supervisor_permissions()
                ):
                    project_ids = [
                        project["id"]
                        for project in user_service.get_projects()
                    ]
                    if permissions.has_supervisor_permissions():
                        department_ids = persons_service.get_current_user(
                            True
                        ).get("departments", [])
                else:
                    raise permissions.PermissionDenied
        try:
            return time_spents_service.get_time_spents(
                person_id,
                date,
                project_ids=project_ids,
                department_ids=department_ids,
            )
        except WrongDateFormatException:
            abort(404)


class DayOffResource(Resource):
    """
    Get day off object for given person and date.
    """

    @jwt_required()
    def get(self, person_id, date):
        """
        Get day off object for given person and date.
        ---
        tags:
        - Persons
        parameters:
          - in: path
            name: person_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: date
            required: True
            type: string
            format: date
            example: "2022-07-12"
        responses:
            200:
                description: Day off object for given person and date
            404:
                description: Wrong date format
        """
        user_service.check_person_is_not_bot(person_id)
        current_user = persons_service.get_current_user()
        if current_user["id"] != person_id:
            try:
                permissions.check_at_least_supervisor_permissions()
            except permissions.PermissionDenied:
                return []
        try:
            return time_spents_service.get_day_off(person_id, date)
        except WrongDateFormatException:
            abort(404)


class PersonDurationTimeSpentsResource(Resource, ArgsMixin):
    """
    Parent class for all person durations time spents resource.
    """

    def get_project_department_arguments(self, person_id):
        project_id = self.get_project_id()
        department_ids = None
        if not permissions.has_admin_permissions():
            if persons_service.get_current_user()["id"] != person_id:
                if (
                    permissions.has_manager_permissions()
                    or permissions.has_supervisor_permissions()
                ):
                    project_ids = [
                        project["id"]
                        for project in user_service.get_projects()
                    ]
                    if project_id is None:
                        project_id = project_ids
                    elif project_id not in project_ids:
                        raise permissions.PermissionDenied
                    if permissions.has_supervisor_permissions():
                        department_ids = persons_service.get_current_user(
                            relations=True
                        )["departments"]
                else:
                    raise permissions.PermissionDenied

        return {
            "project_id": project_id,
            "department_ids": department_ids,
        }


class PersonYearTimeSpentsResource(PersonDurationTimeSpentsResource):
    """
    Get aggregated time spents for given person and year.
    """

    @jwt_required()
    def get(self, person_id, year):
        """
        Get aggregated time spents for given person and year.
        ---
        tags:
        - Persons
        parameters:
          - in: path
            name: person_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: year
            required: True
            type: integer
            example: "2022"
        responses:
            200:
                description: Aggregated time spents for given person and year
            404:
                description: Wrong date format
        """
        user_service.check_person_is_not_bot(person_id)
        try:
            return time_spents_service.get_year_time_spents(
                person_id,
                year,
                **self.get_project_department_arguments(person_id),
            )
        except WrongDateFormatException:
            abort(404)


class PersonMonthTimeSpentsResource(PersonDurationTimeSpentsResource):
    """
    Get aggregated time spents for given person and month.
    """

    @jwt_required()
    def get(self, person_id, year, month):
        """
        Get aggregated time spents for given person and month.
        ---
        tags:
        - Persons
        parameters:
          - in: path
            name: person_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: year
            required: True
            type: integer
            example: "2022"
          - in: path
            name: month
            required: True
            type: integer
            example: 07
            minimum: 1
            maximum: 12
        responses:
            200:
                description: Aggregated time spents for given person and month
            404:
                description: Wrong date format
        """
        user_service.check_person_is_not_bot(person_id)
        try:
            return time_spents_service.get_month_time_spents(
                person_id,
                year,
                month,
                **self.get_project_department_arguments(person_id),
            )
        except WrongDateFormatException:
            abort(404)


class PersonMonthAllTimeSpentsResource(Resource):
    """
    Get all time spents for a given person and month.
    """

    @jwt_required()
    def get(self, person_id, year, month):
        user_service.check_person_is_not_bot(person_id)
        user_service.check_person_access(person_id)
        try:
            timespents = time_spents_service.get_time_spents_for_month(
                year, month, person_id=person_id
            )
            return fields.serialize_list(timespents)
        except WrongDateFormatException:
            abort(404)


class PersonWeekTimeSpentsResource(PersonDurationTimeSpentsResource):
    """
    Get aggregated time spents for given person and week.
    """

    @jwt_required()
    def get(self, person_id, year, week):
        """
        Get aggregated time spents for given person and week.
        ---
        tags:
        - Persons
        parameters:
          - in: path
            name: person_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: year
            required: True
            type: integer
            example: "2022"
          - in: path
            name: week
            required: True
            type: integer
            example: 35
            minimum: 1
            maximum: 52
        responses:
            200:
                description: Aggregated time spents for given person and week
            404:
                description: Wrong date format
        """
        user_service.check_person_is_not_bot(person_id)
        try:
            return time_spents_service.get_week_time_spents(
                person_id,
                year,
                week,
                **self.get_project_department_arguments(person_id),
            )
        except WrongDateFormatException:
            abort(404)


class PersonDayTimeSpentsResource(PersonDurationTimeSpentsResource):
    """
    Get aggregated time spents for given person and day.
    """

    @jwt_required()
    def get(self, person_id, year, month, day):
        """
        Get aggregated time spents for given person and day.
        ---
        tags:
        - Persons
        parameters:
          - in: path
            name: person_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: year
            required: True
            type: integer
            example: "2022"
          - in: path
            name: month
            required: True
            type: integer
            example: 07
            minimum: 1
            maximum: 12
          - in: path
            name: day
            required: True
            type: integer
            example: 12
            minimum: 1
            maximum: 31
        responses:
            200:
                description: Aggregated time spents for given person and day
            404:
                description: Wrong date format
        """
        user_service.check_person_is_not_bot(person_id)
        try:
            return time_spents_service.get_day_time_spents(
                person_id,
                year,
                month,
                day,
                **self.get_project_department_arguments(person_id),
            )
        except WrongDateFormatException:
            abort(404)


class PersonQuotaMixin(ArgsMixin):

    def get_quota_arguments(self):
        project_id = self.get_project_id()
        task_type_id = self.get_task_type_id()
        count_mode = self.get_text_parameter("count_mode", default="weighted")
        if count_mode not in ["weighted", "weighteddone", "feedback", "done"]:
            raise WrongParameterException(
                "count_mode must be equal to weighted, weigtheddone, feedback"
                ", or done"
            )
        feedback = "done" not in count_mode
        weighted = "weighted" in count_mode

        return (project_id, task_type_id, feedback, weighted)

    def check_permissions(self, person_id, project_id=None):
        if permissions.has_manager_permissions():
            user_service.check_manager_project_access(project_id)
        else:
            user_service.check_person_access(person_id)

    def get_person_quotas(self):
        pass

    @jwt_required()
    def get(self, person_id, *args, **kwargs):
        user_service.check_person_is_not_bot(person_id)
        (project_id, task_type_id, feedback, weighted) = (
            self.get_quota_arguments()
        )
        self.check_permissions(person_id, project_id)

        try:
            return self.get_person_quotas(
                person_id,
                *args,
                **kwargs,
                project_id=project_id,
                task_type_id=task_type_id,
                feedback=feedback,
                weighted=weighted,
            )
        except WrongDateFormatException:
            abort(404)


class PersonMonthQuotaShotsResource(Resource, PersonQuotaMixin):
    """
    Get ended shots used for quota calculation of this month.
    """

    def get_person_quotas(self, person_id, year, month, **kwargs):
        return shots_service.get_month_quota_shots(
            person_id, year, month, **kwargs
        )

    @jwt_required()
    def get(self, person_id, year, month):
        """
        Get ended shots used for quota calculation of this month.
        ---
        tags:
        - Persons
        parameters:
          - in: path
            name: person_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: year
            required: True
            type: integer
            example: "2022"
          - in: path
            name: month
            required: True
            type: integer
            example: 07
            minimum: 1
            maximum: 12
          - in: query
            name: count_mode
            required: True
            type: string
            enum: [weighted, weigtheddone, feedback, done]
            example: weighted
        responses:
            200:
                description: Ended shots used for quota calculation of this month
            404:
                description: Wrong date format
        """
        return super().get(person_id, year, month)


class PersonWeekQuotaShotsResource(Resource, PersonQuotaMixin):
    """
    Get ended shots used for quota calculation of this week.
    """

    def get_person_quotas(self, person_id, year, week, **kwargs):
        return shots_service.get_week_quota_shots(
            person_id, year, week, **kwargs
        )

    @jwt_required()
    def get(self, person_id, year, week):
        """
        Get ended shots used for quota calculation of this week.
        ---
        tags:
        - Persons
        parameters:
          - in: path
            name: person_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: year
            required: True
            type: integer
            example: "2022"
          - in: path
            name: week
            required: True
            type: integer
            example: 35
            minimum: 1
            maximum: 52
          - in: query
            name: count_mode
            required: True
            type: string
            enum: [weighted, weigtheddone, feedback, done]
            example: weighted
        responses:
            200:
                description: Ended shots used for quota calculation of this week
            404:
                description: Wrong date format
        """
        return super().get(person_id, year, week)


class PersonDayQuotaShotsResource(Resource, PersonQuotaMixin):
    """
    Get ended shots used for quota calculation of this day.
    """

    def get_person_quotas(self, person_id, year, month, day, **kwargs):
        return shots_service.get_day_quota_shots(
            person_id, year, month, day, **kwargs
        )

    @jwt_required()
    def get(self, person_id, year, month, day):
        """
        Get ended shots used for quota calculation of this day.
        ---
        tags:
        - Persons
        parameters:
          - in: path
            name: person_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: year
            required: True
            type: integer
            example: "2022"
          - in: path
            name: month
            required: True
            type: integer
            example: 07
            minimum: 1
            maximum: 12
          - in: path
            name: day
            required: True
            type: integer
            example: 12
            minimum: 1
            maximum: 31
          - in: query
            name: count_mode
            required: True
            type: string
            enum: [weighted, weigtheddone, feedback, done]
            example: weighted
        responses:
            200:
                description: Ended shots used for quota calculation of this day
            404:
                description: Wrong date format
        """
        return super().get(person_id, year, month, day)


class TimeSpentDurationResource(Resource, ArgsMixin):
    """
    Parent class for all durations time spents resource.
    """

    def get_person_project_department_arguments(self):
        project_id = self.get_project_id()
        person_id = None
        department_id = self.get_text_parameter("department_id")
        if department_id is not None:
            department_ids = [department_id]
        else:
            department_ids = None
        studio_id = self.get_text_parameter("studio_id")
        if not permissions.has_admin_permissions():
            if (
                permissions.has_manager_permissions()
                or permissions.has_supervisor_permissions()
            ):
                project_ids = [
                    project["id"] for project in user_service.get_projects()
                ]
                if project_id is None:
                    project_id = project_ids
                elif project_id not in project_ids:
                    raise permissions.PermissionDenied
                if permissions.has_supervisor_permissions():
                    persons_departments = persons_service.get_current_user(
                        relations=True
                    )["departments"]
                    if department_id is not None:
                        if department_id not in persons_departments:
                            raise WrongParameterException(
                                "Supervisor not allowed to access this department"
                            )
                    else:
                        department_ids = persons_departments
            else:
                person_id = persons_service.get_current_user()["id"]

        return {
            "person_id": person_id,
            "project_id": project_id,
            "department_ids": department_ids,
            "studio_id": studio_id,
        }


class TimeSpentMonthResource(TimeSpentDurationResource):
    """
    Return a table giving time spent by user and by day for given year and
    month.
    """

    @jwt_required()
    def get(self, year, month):
        """
        Return a table giving time spent by user and by day for given year and month.
        ---
        tags:
        - Persons
        parameters:
          - in: path
            name: year
            required: True
            type: integer
            example: "2022"
          - in: path
            name: month
            required: True
            type: integer
            example: 07
            minimum: 1
            maximum: 12
        responses:
            200:
                description: Table giving time spent by user and by day for given year and month
        """

        return time_spents_service.get_day_table(
            year, month, **self.get_person_project_department_arguments()
        )


class TimeSpentYearsResource(TimeSpentDurationResource):
    """
    Return a table giving time spent by user and by month for given year.
    """

    @jwt_required()
    def get(self):
        """
        Return a table giving time spent by user and by month for given year.
        ---
        tags:
        - Persons
        responses:
            200:
                description: Table giving time spent by user and by month for given year
        """
        return time_spents_service.get_year_table(
            **self.get_person_project_department_arguments()
        )


class TimeSpentMonthsResource(TimeSpentDurationResource):
    """
    Return a table giving time spent by user and by month for given year.
    """

    @jwt_required()
    def get(self, year):
        """
        Return a table giving time spent by user and by month for given year.
        ---
        tags:
        - Persons
        parameters:
          - in: path
            name: year
            required: True
            type: integer
            example: "2022"
        responses:
            200:
                description: Table giving time spent by user and by month for given year
        """
        return time_spents_service.get_month_table(
            year, **self.get_person_project_department_arguments()
        )


class TimeSpentWeekResource(TimeSpentDurationResource):
    """
    Return a table giving time spent by user and by week for given year.
    """

    @jwt_required()
    def get(self, year):
        """
        Return a table giving time spent by user and by week for given year.
        ---
        tags:
        - Persons
        parameters:
          - in: path
            name: year
            required: True
            type: integer
            example: "2022"
        responses:
            200:
                description: Table giving time spent by user and by week for given year
        """
        return time_spents_service.get_week_table(
            year, **self.get_person_project_department_arguments()
        )


class InvitePersonResource(Resource):
    """
    Sends an email to given person to invite him/her to connect to Kitsu.
    """

    @jwt_required()
    def get(self, person_id):
        """
        Sends an email to given person to invite him/her to connect to Kitsu.
        ---
        tags:
        - Persons
        parameters:
          - in: path
            name: person_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: Email sent
        """
        user_service.check_person_is_not_bot(person_id)
        permissions.check_admin_permissions()
        persons_service.invite_person(person_id)
        return {"success": True, "message": "Email sent"}


class DayOffForMonthResource(Resource, ArgsMixin):
    """
    Return all day off recorded for given month.
    """

    @jwt_required()
    def get(self, year, month):
        """
        Return all day off recorded for given month.
        ---
        tags:
        - Persons
        parameters:
          - in: path
            name: year
            required: True
            type: integer
            example: "2022"
          - in: path
            name: month
            required: True
            type: integer
            example: 07
            minimum: 1
            maximum: 12
        responses:
            200:
                description: All day off recorded for given month
        """
        if permissions.has_admin_permissions():
            return time_spents_service.get_day_offs_for_month(year, month)
        else:
            person_id = persons_service.get_current_user()["id"]
            return time_spents_service.get_person_day_offs_for_month(
                person_id, year, month
            )


class PersonWeekDayOffResource(Resource, ArgsMixin):
    """
    Return all day off recorded for given week and person.
    """

    @jwt_required()
    def get(self, person_id, year, week):
        """
        Return all day off recorded for given week and person.
        ---
        tags:
        - Persons
        parameters:
          - in: path
            name: person_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: year
            required: True
            type: integer
            example: "2022"
          - in: path
            name: week
            required: True
            type: integer
            example: 35
            minimum: 1
            maximum: 52
        responses:
            200:
                description: All day off recorded for given week and person
        """
        user_service.check_person_is_not_bot(person_id)
        user_service.check_person_access(person_id)
        return time_spents_service.get_person_day_offs_for_week(
            person_id, year, week
        )


class PersonMonthDayOffResource(Resource, ArgsMixin):
    """
    Return all day off recorded for given month and person.
    """

    @jwt_required()
    def get(self, person_id, year, month):
        """
        Return all day off recorded for given month and person.
        ---
        tags:
        - Persons
        parameters:
          - in: path
            name: person_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: year
            required: True
            type: integer
            example: "2022"
          - in: path
            name: month
            required: True
            type: integer
            example: 07
            minimum: 1
            maximum: 12
        responses:
            200:
                description: All day off recorded for given month and person
        """
        user_service.check_person_is_not_bot(person_id)
        user_service.check_person_access(person_id)
        return time_spents_service.get_person_day_offs_for_month(
            person_id, year, month
        )


class PersonYearDayOffResource(Resource, ArgsMixin):
    """
    Return all day off recorded for given year and person.
    """

    @jwt_required()
    def get(self, person_id, year):
        """
        Return all day off recorded for given year and person.
        ---
        tags:
        - Persons
        parameters:
          - in: path
            name: person_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: year
            required: True
            type: integer
            example: "2022"
        responses:
            200:
                description: All day off recorded for given year and person
        """
        user_service.check_person_is_not_bot(person_id)
        user_service.check_person_access(person_id)
        return time_spents_service.get_person_day_offs_for_year(
            person_id, year
        )


class PersonDayOffResource(Resource, ArgsMixin):
    """
    Return all day offs recorded for given and person.
    """

    @jwt_required()
    def get(self, person_id):
        """
        Return all day offs recorded for given and person.
        ---
        tags:
        - Persons
        parameters:
          - in: path
            name: person_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All day off recorded for given person.
        """
        user_service.check_person_is_not_bot(person_id)
        user_service.check_person_access(person_id)
        return time_spents_service.get_day_offs_between(
            person_id=person_id,
        )


class AddToDepartmentResource(Resource, ArgsMixin):
    """
    Add a user to given department.
    """

    @jwt_required()
    def post(self, person_id):
        """
        Add a user to given department.
        ---
        tags:
        - Persons
        parameters:
          - in: path
            name: person_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            201:
                description: User added to given department
        """
        args = self.get_args(
            [
                ("department_id", None, True),
            ]
        )

        permissions.check_admin_permissions()

        try:
            department = tasks_service.get_department(args["department_id"])
        except DepartmentNotFoundException:
            raise WrongParameterException(
                "Department ID matches no department"
            )
        person = persons_service.add_to_department(department["id"], person_id)
        return person, 201


class RemoveFromDepartmentResource(Resource, ArgsMixin):
    """
    Remove a user from given department.
    """

    @jwt_required()
    def delete(self, person_id, department_id):
        """
        Remove a user from given department.
        ---
        tags:
        - Persons
        parameters:
          - in: path
            name: person_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: department_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            204:
                description: User removed from given department
        """
        permissions.check_admin_permissions()
        try:
            tasks_service.get_department(department_id)
        except DepartmentNotFoundException:
            raise WrongParameterException(
                "Department ID matches no department"
            )
        persons_service.remove_from_department(department_id, person_id)
        return "", 204


class ChangePasswordForPersonResource(Resource, ArgsMixin):
    """
    Allow admin to change password for given user.
    """

    @jwt_required()
    def post(self, person_id):
        """
        Allow admin to change password for given user.
        ---
        description: Prior to modifying the password, it requires to be admin.
                     An admin can't change other admins password.
                     The new password requires a confirmation to ensure that the admin didn't
                     make a mistake by typing the new password.
        tags:
            - Persons
        parameters:
          - in: path
            name: person_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: formData
            name: password
            required: True
            type: string
            format: password
          - in: formData
            name: password_2
            required: True
            type: string
            format: password
        responses:
          200:
            description: Password changed
          400:
            description: Invalid password or inactive user
        """
        user_service.check_person_is_not_bot(person_id)
        (password, password_2) = self.get_arguments()
        permissions.check_admin_permissions()
        try:
            person = persons_service.get_person(person_id)
            if (
                person["email"] in config.PROTECTED_ACCOUNTS
                and person["id"] != persons_service.get_current_user()["id"]
            ):
                raise PersonInProtectedAccounts()
            current_user = persons_service.get_current_user()
            auth.validate_password(password, password_2)
            password = auth.encrypt_password(password)
            persons_service.update_password(person["email"], password)
            current_app.logger.warning(
                "User %s has changed the password of %s"
                % (current_user["email"], person["email"])
            )
            organisation = persons_service.get_organisation()
            time_string = format_datetime(
                date_helpers.get_utc_now_datetime(),
                tzinfo=person["timezone"],
                locale=person["locale"],
            )
            person_IP = request.headers.get("X-Forwarded-For", None)
            html = f"""<p>Hello {person["first_name"]},</p>
<p>
Your password was changed at this date: {time_string}.
The IP of the user who changed your password is: {person_IP}.
If you don't know the person who changed the password, please contact our support team.
</p>
Thank you and see you soon on Kitsu,
</p>
<p>
{organisation["name"]} Team
</p>
"""
            subject = f"{organisation['name']} - Kitsu: password changed"
            emails.send_email(subject, html, person["email"])
            return {"success": True}

        except auth.PasswordsNoMatchException:
            return (
                {
                    "error": True,
                    "message": "Confirmation password doesn't match.",
                },
                400,
            )
        except auth.PasswordTooShortException:
            return {"error": True, "message": "Password is too short."}, 400
        except UnactiveUserException:
            return {"error": True, "message": "User is unactive."}, 400
        except PersonInProtectedAccounts:
            return (
                {
                    "error": True,
                    "message": "This user is in protected accounts.",
                },
                400,
            )

    def get_arguments(self):
        args = self.get_args(
            [
                {
                    "name": "password",
                    "required": True,
                    "help": "New password is missing.",
                },
                {
                    "name": "password_2",
                    "required": True,
                    "help": "New password confirmation is missing.",
                },
            ]
        )

        return (args["password"], args["password_2"])


class DisableTwoFactorAuthenticationPersonResource(Resource, ArgsMixin):
    """
    Allow admin to disable two factor authentication for given user.
    """

    @jwt_required()
    def delete(self, person_id):
        """
        Allow admin to disable two factor authentication for given user.
        ---
        description: Prior to disable two factor authentication, it requires to
                     be admin.
                     An admin can't disable two factor authentication for other
                     admins.
        tags:
            - Persons
        parameters:
          - in: path
            name: person_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
          200:
            description: Two factor authentication disabled
          400:
            description: Inactive user
        """
        user_service.check_person_is_not_bot(person_id)
        permissions.check_admin_permissions()
        try:
            person = persons_service.get_person(person_id)
            current_user = persons_service.get_current_user()
            disable_two_factor_authentication_for_person(person["id"])
            current_app.logger.warning(
                "User %s has disabled the two factor authentication of %s"
                % (current_user["email"], person["email"])
            )
            organisation = persons_service.get_organisation()
            time_string = format_datetime(
                date_helpers.get_utc_now_datetime(),
                tzinfo=person["timezone"],
                locale=person["locale"],
            )
            person_IP = request.headers.get("X-Forwarded-For", None)
            html = f"""<p>Hello {person["first_name"]},</p>
<p>
Your two factor authentication was disabled at this date: {time_string}.
The IP of the user who disabled your two factor authentication is: {person_IP}.
If you don't know the person who disabled the two factor authentication, please contact our support team.
</p>
Thank you and see you soon on Kitsu,
</p>
<p>
{organisation["name"]} Team
</p>
"""
            subject = f"{organisation['name']} - Kitsu: two factor authentication disabled"
            emails.send_email(subject, html, person["email"])
            return {"success": True}

        except UnactiveUserException:
            return {"error": True, "message": "User is unactive."}, 400
        except TwoFactorAuthenticationNotEnabledException:
            return {
                "error": True,
                "message": "Two factor authentication not enabled for this user.",
            }, 400


class ClearAvatarPersonResource(Resource):
    @jwt_required()
    def delete(self, person_id):
        """
        Set `has_avatar` flag to False for current user and remove its avatar
        file.
        ---
        tags:
          - User
        responses:
            204:
                description: Avatar file deleted
        """
        permissions.check_admin_permissions()
        persons_service.clear_avatar(person_id)
        return "", 204



"""ProductionScheduleVersion add on delete cascade to fkey

Revision ID: 4368137b44e1
Revises: 0bd1e89f2a6f
Create Date: 2025-07-23 00:41:37.354903

"""

from alembic import op


# revision identifiers, used by Alembic.
revision = "4368137b44e1"
down_revision = "0bd1e89f2a6f"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table(
        "production_schedule_version", schema=None
    ) as batch_op:
        batch_op.drop_constraint(
            batch_op.f(
                "production_schedule_version_production_schedule_from_fkey"
            ),
            type_="foreignkey",
        )
        batch_op.drop_constraint(
            batch_op.f("production_schedule_version_project_id_fkey"),
            type_="foreignkey",
        )
        batch_op.create_foreign_key(
            None, "project", ["project_id"], ["id"], ondelete="CASCADE"
        )
        batch_op.create_foreign_key(
            None,
            "production_schedule_version",
            ["production_schedule_from"],
            ["id"],
            ondelete="CASCADE",
        )

    with op.batch_alter_table(
        "production_schedule_version_task_link", schema=None
    ) as batch_op:
        batch_op.drop_constraint(
            batch_op.f(
                "production_schedule_version_t_production_schedule_version__fkey"
            ),
            type_="foreignkey",
        )
        batch_op.drop_constraint(
            batch_op.f("production_schedule_version_task_link_task_id_fkey"),
            type_="foreignkey",
        )
        batch_op.create_foreign_key(
            None,
            "production_schedule_version",
            ["production_schedule_version_id"],
            ["id"],
            ondelete="CASCADE",
        )
        batch_op.create_foreign_key(
            None, "task", ["task_id"], ["id"], ondelete="CASCADE"
        )

    with op.batch_alter_table(
        "production_schedule_version_task_link_person_link", schema=None
    ) as batch_op:
        batch_op.drop_constraint(
            batch_op.f(
                "production_schedule_version_task_link_person_lin_person_id_fkey"
            ),
            type_="foreignkey",
        )
        batch_op.drop_constraint(
            batch_op.f(
                "production_schedule_version__production_schedule_version__fkey1"
            ),
            type_="foreignkey",
        )
        batch_op.create_foreign_key(
            None,
            "production_schedule_version_task_link",
            ["production_schedule_version_task_link_id"],
            ["id"],
            ondelete="CASCADE",
        )
        batch_op.create_foreign_key(
            None, "person", ["person_id"], ["id"], ondelete="CASCADE"
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table(
        "production_schedule_version_task_link_person_link", schema=None
    ) as batch_op:
        batch_op.drop_constraint(
            batch_op.f(
                "production_schedule_version_task_link_person_lin_person_id_fkey"
            ),
            type_="foreignkey",
        )
        batch_op.drop_constraint(
            batch_op.f(
                "production_schedule_version__production_schedule_version__fkey1"
            ),
            type_="foreignkey",
        )
        batch_op.create_foreign_key(
            batch_op.f(
                "production_schedule_version__production_schedule_version__fkey1"
            ),
            "production_schedule_version_task_link",
            ["production_schedule_version_task_link_id"],
            ["id"],
        )
        batch_op.create_foreign_key(
            batch_op.f(
                "production_schedule_version_task_link_person_lin_person_id_fkey"
            ),
            "person",
            ["person_id"],
            ["id"],
        )

    with op.batch_alter_table(
        "production_schedule_version_task_link", schema=None
    ) as batch_op:
        batch_op.drop_constraint(
            batch_op.f(
                "production_schedule_version_t_production_schedule_version__fkey"
            ),
            type_="foreignkey",
        )
        batch_op.drop_constraint(
            batch_op.f("production_schedule_version_task_link_task_id_fkey"),
            type_="foreignkey",
        )
        batch_op.create_foreign_key(
            batch_op.f("production_schedule_version_task_link_task_id_fkey"),
            "task",
            ["task_id"],
            ["id"],
        )
        batch_op.create_foreign_key(
            batch_op.f(
                "production_schedule_version_t_production_schedule_version__fkey"
            ),
            "production_schedule_version",
            ["production_schedule_version_id"],
            ["id"],
        )

    with op.batch_alter_table(
        "production_schedule_version", schema=None
    ) as batch_op:
        batch_op.drop_constraint(
            batch_op.f(
                "production_schedule_version_production_schedule_from_fkey"
            ),
            type_="foreignkey",
        )
        batch_op.drop_constraint(
            batch_op.f("production_schedule_version_project_id_fkey"),
            type_="foreignkey",
        )
        batch_op.create_foreign_key(
            batch_op.f("production_schedule_version_project_id_fkey"),
            "project",
            ["project_id"],
            ["id"],
        )
        batch_op.create_foreign_key(
            batch_op.f(
                "production_schedule_version_production_schedule_from_fkey"
            ),
            "production_schedule_version",
            ["production_schedule_from"],
            ["id"],
        )

    # ### end Alembic commands ###



from zou.app.models.production_schedule_version import (
    ProductionScheduleVersion,
    ProductionScheduleVersionTaskLink,
)

from zou.app.blueprints.crud.base import BaseModelsResource, BaseModelResource
from zou.app.services import user_service, tasks_service, schedule_service
from zou.app.utils import permissions
from zou.app.services.exception import WrongParameterException


class ProductionScheduleVersionsResource(BaseModelsResource):
    def __init__(self):
        BaseModelsResource.__init__(self, ProductionScheduleVersion)

    def check_read_permissions(self, options=None):
        if (
            permissions.has_vendor_permissions()
            or permissions.has_client_permissions()
        ):
            raise permissions.PermissionDenied
        if "project_id" in options.keys():
            return user_service.check_project_access(options["project_id"])
        else:
            return permissions.check_admin_permissions()

    def check_create_permissions(self, data):
        return user_service.check_manager_project_access(
            project_id=data["project_id"]
        )


class ProductionScheduleVersionResource(BaseModelResource):
    def __init__(self):
        BaseModelResource.__init__(self, ProductionScheduleVersion)

    def check_read_permissions(self, instance_dict):
        if (
            permissions.has_vendor_permissions()
            or permissions.has_client_permissions()
        ):
            raise permissions.PermissionDenied
        return user_service.check_project_access(instance_dict["project_id"])

    def check_update_permissions(self, instance_dict, data):
        return user_service.check_manager_project_access(
            project_id=instance_dict["project_id"]
        )


class ProductionScheduleVersionTaskLinksResource(BaseModelsResource):
    def __init__(self):
        BaseModelsResource.__init__(self, ProductionScheduleVersionTaskLink)

    def check_read_permissions(self, options=None):
        if (
            permissions.has_vendor_permissions()
            or permissions.has_client_permissions()
        ):
            raise permissions.PermissionDenied
        if "project_id" in options.keys():
            return user_service.check_project_access(options["project_id"])
        else:
            return permissions.check_admin_permissions()

    def check_create_permissions(self, data):
        project_id_from_production_version_schedule = (
            schedule_service.get_production_schedule_version(
                data["production_schedule_version_id"]
            )["project_id"]
        )
        project_id_from_task = tasks_service.get_task(data["task_id"])[
            "project_id"
        ]
        if project_id_from_production_version_schedule != project_id_from_task:
            raise WrongParameterException(
                "The task and the production schedule version must be in the same project."
            )
        return user_service.check_manager_project_access(
            project_id=project_id_from_production_version_schedule
        )


class ProductionScheduleVersionTaskLinkResource(BaseModelResource):
    def __init__(self):
        BaseModelResource.__init__(self, ProductionScheduleVersionTaskLink)
        self.protected_fields = [
            "id",
            "created_at",
            "updated_at",
            "task_id",
            "production_schedule_version_id",
        ]

    def check_read_permissions(self, instance_dict):
        if (
            permissions.has_vendor_permissions()
            or permissions.has_client_permissions()
        ):
            raise permissions.PermissionDenied
        task = tasks_service.get_task(instance_dict["task_id"])
        return user_service.check_project_access(task["project_id"])

    def check_update_permissions(self, instance_dict, data):
        task = tasks_service.get_task(instance_dict["task_id"])
        return user_service.check_manager_project_access(
            project_id=task["project_id"]
        )

    def check_delete_permissions(self, instance_dict):
        task = tasks_service.get_task(instance_dict["task_id"])
        return user_service.check_manager_project_access(
            project_id=task["project_id"]
        )



from zou.app.models.organisation import Organisation
from zou.app.blueprints.crud.base import BaseModelResource, BaseModelsResource

from zou.app.services import persons_service
from zou.app.utils.permissions import has_admin_permissions


class OrganisationsResource(BaseModelsResource):
    def __init__(self):
        BaseModelsResource.__init__(self, Organisation)

    def check_read_permissions(self, options=None):
        return True


class OrganisationResource(BaseModelResource):
    def __init__(self):
        BaseModelResource.__init__(self, Organisation)

    def check_read_permissions(self, instance):
        return True

    def pre_update(self, instance_dict, data):
        if "hours_by_day" in data:
            data["hours_by_day"] = float(data["hours_by_day"])
        return data

    def serialize_instance(self, data, relations=True):
        return data.serialize(
            relations=relations,
            ignored_attrs=(
                []
                if has_admin_permissions()
                else [
                    "chat_token_slack",
                    "chat_webhook_mattermost",
                    "chat_token_discord",
                ]
            ),
        )

    def post_update(self, instance_dict, data):
        persons_service.clear_organisation_cache()
        return instance_dict



import flask_bcrypt
import email_validator

from zou.app import config


class PasswordTooShortException(BaseException):
    pass


class PasswordsNoMatchException(BaseException):
    pass


class EmailNotValidException(BaseException):
    pass


def encrypt_password(password):
    """
    Encrypt given string password using bcrypt algorithm.
    """
    return flask_bcrypt.generate_password_hash(password)


def validate_email(
    email, check_deliverability=config.MAIL_CHECK_DELIVERABILITY
):
    try:
        return email_validator.validate_email(
            email, check_deliverability=check_deliverability
        ).normalized
    except email_validator.EmailNotValidError as e:
        raise EmailNotValidException(str(e))


def validate_password(password, password_2=None):
    if len(password) < config.MIN_PASSWORD_LENGTH:
        raise PasswordTooShortException()
    if password_2 is not None and password != password_2:
        raise PasswordsNoMatchException()
    return True



from zou.app.models.budget import Budget
from zou.app.models.budget_entry import BudgetEntry

from zou.app.services.exception import (
    BudgetNotFoundException,
    BudgetEntryNotFoundException,
)

from zou.app.utils import events
from zou.app.utils import fields

from sqlalchemy.exc import StatementError


def get_budget_raw(budget_id):
    """
    Return budget corresponding to given budget ID.
    """
    if budget_id is None:
        raise BudgetNotFoundException()

    try:
        budget = Budget.get(budget_id)
    except StatementError:
        raise BudgetNotFoundException()

    if budget is None:
        raise BudgetNotFoundException()
    return budget


def get_budget(budget_id):
    """
    Return budget corresponding to given budget ID as a dictionary.
    """
    return get_budget_raw(budget_id).serialize(relations=True)


def get_budgets(project_id):
    """
    Return all budgets for given project ID.
    """
    budgets = Budget.get_all_by(project_id=project_id)
    return fields.present_models(budgets)


def create_budget(project_id, name, currency=None):
    """
    Create a new budget for given project ID.
    """
    last_budget = (
        Budget.query.filter_by(project_id=project_id)
        .order_by(Budget.revision.desc())
        .first()
    )
    last_revision = 1
    if last_budget is not None:
        last_revision = last_budget.revision + 1
    budget = Budget(
        project_id=project_id,
        name=name,
        currency=currency,
        revision=last_revision,
    )
    budget.save()
    events.emit(
        "budget:create",
        {"budget_id": str(budget.id)},
        project_id=project_id,
    )
    return budget.serialize()


def update_budget(budget_id, name=None, currency=None):
    """
    Update budget corresponding to given budget ID.
    """
    budget = get_budget_raw(budget_id)
    if name is not None:
        budget.name = name
    if currency is not None:
        budget.currency = currency
    budget.save()
    events.emit(
        "budget:update",
        {"budget_id": str(budget.id)},
        project_id=str(budget.project_id),
    )
    return budget.serialize()


def delete_budget(budget_id):
    """
    Delete budget corresponding to given budget ID.
    """
    budget = get_budget_raw(budget_id)
    budget_entries = BudgetEntry.delete_all_by(budget_id=budget_id)
    budget.delete()
    events.emit(
        "budget:delete",
        {"budget_id": budget_id},
        project_id=str(budget.project_id),
    )
    return budget.serialize()


def get_budget_entries(budget_id):
    """
    Return all budget entries for given budget ID.
    """
    budget_entries = BudgetEntry.get_all_by(budget_id=budget_id)
    return fields.present_models(budget_entries)


def get_budget_entry_raw(budget_entry_id):
    """
    Return budget entry corresponding to given budget entry ID.
    """
    try:
        budget_entry = BudgetEntry.get(budget_entry_id)
    except StatementError:
        raise BudgetEntryNotFoundException()

    if budget_entry is None:
        raise BudgetEntryNotFoundException()
    return budget_entry


def get_budget_entry(budget_entry_id):
    """
    Return budget entry corresponding to given budget entry ID as a dictionary.
    """
    return get_budget_entry_raw(budget_entry_id).serialize()


def create_budget_entry(
    budget_id,
    department_id,
    start_date,
    months_duration,
    daily_salary,
    position,
    seniority,
    person_id=None,
):
    """
    Create a new budget entry for given budget ID.
    """
    budget = get_budget_raw(budget_id)
    budget_entry = BudgetEntry.create(
        budget_id=budget_id,
        department_id=department_id,
        person_id=person_id,
        start_date=start_date,
        months_duration=months_duration,
        daily_salary=daily_salary,
        position=position,
        seniority=seniority,
        exceptions={},
    )
    events.emit(
        "budget-entry:create",
        {"budget_id": str(budget_id), "budget_entry_id": str(budget_entry.id)},
        project_id=str(budget.project_id),
    )
    return budget_entry.serialize()


def update_budget_entry(budget_entry_id, data):
    """
    Update budget entry corresponding to given budget entry ID.
    """
    budget_entry = get_budget_entry_raw(budget_entry_id)
    budget = get_budget_raw(str(budget_entry.budget_id))
    data = _clean_exceptions(data)
    budget_entry.update(data)
    events.emit(
        "budget-entry:update",
        {"budget_id": str(budget.id), "budget_entry_id": str(budget_entry.id)},
        project_id=str(budget.project_id),
    )
    return budget_entry.serialize()


def _clean_exceptions(data):
    if "exceptions" in data:
        if data["exceptions"] is None:
            data["exceptions"] = {}
        else:
            data["exceptions"] = {
                k: int(v)
                for k, v in data["exceptions"].items()
                if v is not None and int(v) > 0
            }
    return data


def delete_budget_entry(budget_entry_id):
    """
    Delete budget entry corresponding to given budget entry ID.
    """
    budget_entry = get_budget_entry_raw(budget_entry_id)
    budget = get_budget_raw(str(budget_entry.budget_id))
    budget_entry.delete()
    events.emit(
        "budget-entry:delete",
        {"budget_id": str(budget.id), "budget_entry_id": budget_entry_id},
        project_id=str(budget.project_id),
    )
    return budget_entry.serialize()



from flask import request
from flask_restful import Resource
from flask_jwt_extended import jwt_required

from zou.app.services import (
    persons_service,
    projects_service,
    playlists_service,
    edits_service,
    tasks_service,
    user_service,
)

from zou.app.mixin import ArgsMixin
from zou.app.utils import permissions, query


class EditResource(Resource, ArgsMixin):
    @jwt_required()
    def get(self, edit_id):
        """
        Retrieve given edit.
        ---
        tags:
          - Edits
        parameters:
          - in: path
            name: edit_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: Given edit
        """
        edit = edits_service.get_full_edit(edit_id)
        if edit is None:
            edits_service.clear_edit_cache(edit_id)
            edit = edits_service.get_full_edit(edit_id)
        user_service.check_project_access(edit["project_id"])
        user_service.check_entity_access(edit["id"])
        return edit

    @jwt_required()
    def delete(self, edit_id):
        """
        Delete given edit.
        ---
        tags:
          - Edits
        parameters:
          - in: path
            name: edit_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            204:
                description: Given edit deleted
        """
        force = self.get_force()
        edit = edits_service.get_edit(edit_id)
        if edit["created_by"] == persons_service.get_current_user()["id"]:
            user_service.check_belong_to_project(edit["project_id"])
        else:
            user_service.check_manager_project_access(edit["project_id"])
        edits_service.remove_edit(edit_id, force=force)
        return "", 204


class EditsResource(Resource):
    @jwt_required()
    def get(self):
        """
        Retrieve all edit entries.
        ---
        tags:
          - Edits
        description: Filters can be specified in the query string.
        parameters:
          - in: query
            name: project_id
            required: False
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: query
            name: name
            required: False
            type: string
            example: Name of edit
          - in: query
            name: force
            required: False
            type: boolean
            default: False
        responses:
            200:
                description: All edit entries
        """
        criterions = query.get_query_criterions_from_request(request)
        user_service.check_project_access(criterions.get("project_id", None))
        if permissions.has_vendor_permissions():
            criterions["assigned_to"] = persons_service.get_current_user()[
                "id"
            ]
        return edits_service.get_edits(criterions)


class AllEditsResource(Resource):
    @jwt_required()
    def get(self):
        """
        Retrieve all edit entries.
        ---
        tags:
          - Edits
        description: Filters can be specified in the query string.
        parameters:
          - in: query
            name: project_id
            required: False
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: query
            name: name
            required: False
            type: string
            example: Name of edit
          - in: query
            name: force
            required: False
            type: boolean
            default: False
        responses:
            200:
                description: All edit entries
        """
        criterions = query.get_query_criterions_from_request(request)
        if permissions.has_vendor_permissions():
            criterions["assigned_to"] = persons_service.get_current_user()[
                "id"
            ]
        user_service.check_project_access(criterions.get("project_id", None))
        return edits_service.get_edits(criterions)


class EditTaskTypesResource(Resource):
    @jwt_required()
    def get(self, edit_id):
        """
        Retrieve all task types related to a given edit.
        ---
        tags:
          - Edits
        parameters:
          - in: path
            name: edit_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All task types related to given edit
        """
        edit = edits_service.get_edit(edit_id)
        user_service.check_project_access(edit["project_id"])
        user_service.check_entity_access(edit["id"])
        return tasks_service.get_task_types_for_edit(edit_id)


class EditTasksResource(Resource, ArgsMixin):
    @jwt_required()
    def get(self, edit_id):
        """
        Retrieve all tasks related to a given edit.
        ---
        tags:
          - Edits
        parameters:
          - in: path
            name: edit_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All tasks related to given edit
        """
        edit = edits_service.get_edit(edit_id)
        user_service.check_project_access(edit["project_id"])
        user_service.check_entity_access(edit["id"])
        relations = self.get_relations()
        return tasks_service.get_tasks_for_edit(edit_id, relations=relations)


class EpisodeEditTasksResource(Resource, ArgsMixin):
    @jwt_required()
    def get(self, episode_id):
        """
        Retrieve all tasks related to a given episode.
        ---
        tags:
          - Edits
        parameters:
          - in: path
            name: episode_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All tasks related to given episode
        """
        episode = edits_service.get_episode(episode_id)
        user_service.check_project_access(episode["project_id"])
        user_service.check_entity_access(episode["id"])
        if permissions.has_vendor_permissions():
            raise permissions.PermissionDenied
        relations = self.get_relations()
        return tasks_service.get_edit_tasks_for_episode(
            episode_id, relations=relations
        )


class EpisodeEditsResource(Resource, ArgsMixin):
    @jwt_required()
    def get(self, episode_id):
        """
        Retrieve all edits related to a given episode.
        ---
        tags:
          - Edits
        parameters:
          - in: path
            name: episode_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All efits related to given episode
        """
        episode = edits_service.get_episode(episode_id)
        user_service.check_project_access(episode["project_id"])
        user_service.check_entity_access(episode["id"])
        relations = self.get_relations()
        return edits_service.get_edits_for_episode(
            episode_id, relations=relations
        )


class EditPreviewsResource(Resource):
    @jwt_required()
    def get(self, edit_id):
        """
        Retrieve all previews related to a given edit.
        ---
        tags:
          - Edits
        description: It sends them as a dict. Keys are related task type ids and values are arrays of preview for this task type.
        parameters:
          - in: path
            name: edit_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All previews related to given edit
        """
        edit = edits_service.get_edit(edit_id)
        user_service.check_project_access(edit["project_id"])
        user_service.check_entity_access(edit["id"])
        return playlists_service.get_preview_files_for_entity(edit_id)


class EditsAndTasksResource(Resource):
    @jwt_required()
    def get(self):
        """
        Retrieve all edits, adds project name and all related tasks.
        ---
        tags:
          - Edits
        parameters:
          - in: query
            name: project_id
            required: False
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: query
            name: name
            required: False
            type: string
            example: Name of edit
          - in: query
            name: force
            required: False
            type: boolean
            default: False
        responses:
            200:
                description: All edits and all related tasks.
        """
        criterions = query.get_query_criterions_from_request(request)
        user_service.check_project_access(criterions.get("project_id", None))
        if permissions.has_vendor_permissions():
            criterions["assigned_to"] = persons_service.get_current_user()[
                "id"
            ]
            criterions["vendor_departments"] = [
                str(department.id)
                for department in persons_service.get_current_user_raw().departments
            ]
        return edits_service.get_edits_and_tasks(criterions)


class ProjectEditsResource(Resource, ArgsMixin):
    @jwt_required()
    def get(self, project_id):
        """
        Retrieve all edits related to a given project.
        ---
        tags:
          - Edits
        parameters:
          - in: path
            name: project_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All edits related to given project
        """
        projects_service.get_project(project_id)
        user_service.check_project_access(project_id)
        return edits_service.get_edits_for_project(
            project_id, only_assigned=permissions.has_vendor_permissions()
        )

    @jwt_required()
    def post(self, project_id):
        """
        Create an edit for given project.
        ---
        tags:
          - Edits
        parameters:
          - in: path
            name: project_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: formData
            name: name
            required: True
            type: string
            example: Name of edit
          - in: formData
            name: description
            type: string
            example: Description of edit
          - in: formData
            name: episode_id
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            201:
                description: Edit created for given project
        """
        (name, description, data, parent_id) = self.get_arguments()
        projects_service.get_project(project_id)
        user_service.check_manager_project_access(project_id)

        edit = edits_service.create_edit(
            project_id,
            name,
            data=data,
            description=description,
            parent_id=parent_id,
            created_by=persons_service.get_current_user()["id"],
        )
        return edit, 201

    def get_arguments(self):
        args = self.get_args(
            [
                {
                    "name": "name",
                    "help": "The edit name is required.",
                    "required": True,
                },
                "description",
                {"name": "data", "type": dict},
                "episode_id",
            ]
        )

        return (
            args["name"],
            args.get("description", ""),
            args["data"],
            args["episode_id"],
        )


class EditVersionsResource(Resource):
    """
    Retrieve data versions of given edit.
    """

    @jwt_required()
    def get(self, edit_id):
        """
        Retrieve data versions of given edit.
        ---
        tags:
          - Edits
        parameters:
          - in: path
            name: edit_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: Data versions of given edit
        """
        edit = edits_service.get_edit(edit_id)
        user_service.check_project_access(edit["project_id"])
        user_service.check_entity_access(edit["id"])
        return edits_service.get_edit_versions(edit_id)



"""Add ProjectTaskStatusLink.roles_for_board

Revision ID: 20dfeb36142b
Revises: 5798d2c9020b
Create Date: 2024-01-12 17:22:30.688955

"""

from alembic import op
import sqlalchemy as sa
import sqlalchemy_utils


# revision identifiers, used by Alembic.
revision = "20dfeb36142b"
down_revision = "5798d2c9020b"
branch_labels = None
depends_on = None

ROLE_TYPES = [
    ("user", "Artist"),
    ("admin", "Studio Manager"),
    ("supervisor", "Supervisor"),
    ("manager", "Production Manager"),
    ("client", "Client"),
    ("vendor", "Vendor"),
]


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table(
        "project_task_status_link", schema=None
    ) as batch_op:
        batch_op.add_column(
            sa.Column(
                "roles_for_board",
                sa.ARRAY(sqlalchemy_utils.types.choice.ChoiceType(ROLE_TYPES)),
                nullable=True,
            )
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table(
        "project_task_status_link", schema=None
    ) as batch_op:
        batch_op.drop_column("roles_for_board")

    # ### end Alembic commands ###



"""Remove on delete cascade for ProductionScheduleVersion.production_schedule_from

Revision ID: e8bc24998b34
Revises: 4368137b44e1
Create Date: 2025-07-24 14:37:53.406897

"""

from alembic import op


# revision identifiers, used by Alembic.
revision = "e8bc24998b34"
down_revision = "4368137b44e1"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table(
        "production_schedule_version", schema=None
    ) as batch_op:
        batch_op.drop_constraint(
            batch_op.f(
                "production_schedule_version_production_schedule_from_fkey"
            ),
            type_="foreignkey",
        )
        batch_op.create_foreign_key(
            None,
            "production_schedule_version",
            ["production_schedule_from"],
            ["id"],
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table(
        "production_schedule_version", schema=None
    ) as batch_op:
        batch_op.drop_constraint(
            "production_schedule_version_production_schedule_from_fkey",
            type_="foreignkey",
        )
        batch_op.create_foreign_key(
            batch_op.f(
                "production_schedule_version_production_schedule_from_fkey"
            ),
            "production_schedule_version",
            ["production_schedule_from"],
            ["id"],
            ondelete="CASCADE",
        )

    # ### end Alembic commands ###



"""Add column estimation for ProductionScheduleVersionTaskLink

Revision ID: ce7f46f445dc
Revises: e8bc24998b34
Create Date: 2025-07-24 19:27:55.423759

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "ce7f46f445dc"
down_revision = "e8bc24998b34"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table(
        "production_schedule_version_task_link", schema=None
    ) as batch_op:
        batch_op.add_column(sa.Column("estimation", sa.Float(), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table(
        "production_schedule_version_task_link", schema=None
    ) as batch_op:
        batch_op.drop_column("estimation")

    # ### end Alembic commands ###



from flask_restful import Resource
from flask_jwt_extended import jwt_required
from slugify import slugify
from sqlalchemy.orm import aliased

from zou.app.models.entity import Entity, EntityLink
from zou.app.models.entity_type import EntityType

from zou.app.services import projects_service, shots_service, user_service
from zou.app.utils import csv_utils

from zou.app.mixin import ArgsMixin


class CastingCsvExport(Resource, ArgsMixin):
    @jwt_required()
    def get(self, project_id):
        """
        Export casting linked to a given project as csv.
        ---
        tags:
          - Export
        parameters:
          - in: path
            name: project_id
            required: True
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: Casting exported as csv
        """
        project = projects_service.get_project(project_id)  # Check existence
        self.check_permissions(project_id)

        episode_id = self.get_episode_id()

        is_shot_casting = self.get_bool_parameter("is_shot_casting")

        is_tv_show = projects_service.is_tv_show(project)

        results = self.build_results(
            project_id,
            is_tv_show=is_tv_show,
            episode_id=episode_id,
            is_shot_casting=is_shot_casting,
        )
        headers = self.build_headers(
            is_tv_show=is_tv_show, episode_id=episode_id
        )

        csv_content = [headers]
        for result in results:
            csv_content.append(
                self.build_row(
                    result, is_tv_show=is_tv_show, episode_id=episode_id
                )
            )

        file_name = "%s casting" % project["name"]
        return csv_utils.build_csv_response(csv_content, slugify(file_name))

    def check_permissions(self, project_id):
        user_service.check_project_access(project_id)
        user_service.block_access_to_vendor()

    def build_headers(self, is_tv_show=False, episode_id=None):
        headers = [
            "Parent",
            "Name",
            "Asset Type",
            "Asset",
            "Occurences",
            "Label",
        ]
        if is_tv_show and episode_id not in ["all"]:
            return ["Episode"] + headers
        else:
            return headers

    def build_row(self, result, is_tv_show=False, episode_id=None):
        if is_tv_show and episode_id not in ["all"]:
            (
                episode_name,
                target_parent_name,
                target_entity_type_name,
                target_name,
                asset_type_name,
                asset_name,
                entity_link_nb_occurences,
                entity_link_label,
            ) = result
            row = [
                episode_name or "",
                target_parent_name or target_entity_type_name,
                target_name,
                asset_type_name,
                asset_name,
                entity_link_nb_occurences,
                entity_link_label or "",
            ]
        else:
            (
                target_parent_name,
                target_entity_type_name,
                target_name,
                asset_type_name,
                asset_name,
                entity_link_nb_occurences,
                entity_link_label,
            ) = result
            row = [
                target_parent_name or target_entity_type_name,
                target_name,
                asset_type_name,
                asset_name,
                entity_link_nb_occurences,
                entity_link_label or "",
            ]
        return row

    def build_results(
        self,
        project_id,
        is_tv_show=False,
        episode_id=None,
        is_shot_casting=False,
    ):
        results = []
        if episode_id == "main":
            results = self.build_main_pack_results(project_id)
        elif episode_id == "all":
            results = self.build_episodes_results(project_id)
        elif is_shot_casting:
            results = self.build_shot_results(
                project_id, is_tv_show, episode_id
            )
        else:
            results = self.build_asset_results(
                project_id, is_tv_show, episode_id
            )
        return results

    def build_shot_results(
        self, project_id, is_tv_show=False, episode_id=None
    ):
        results = []
        Shot = aliased(Entity, name="shot")
        Asset = aliased(Entity, name="asset")
        Sequence = aliased(Entity, name="sequence")
        Episode = aliased(Entity, name="episode")
        AssetType = aliased(EntityType, name="asset_type")

        query = (
            EntityLink.query.join(Shot, EntityLink.entity_in_id == Shot.id)
            .join(Sequence, Shot.parent_id == Sequence.id)
            .join(Asset, EntityLink.entity_out_id == Asset.id)
            .join(AssetType, Asset.entity_type_id == AssetType.id)
            .filter(Shot.project_id == project_id)
        )
        if is_tv_show:
            query = query.join(Episode, Sequence.parent_id == Episode.id)
            if episode_id is not None:
                query = query.filter(Episode.id == episode_id)
            query = query.add_columns(
                Episode.name,
                Sequence.name,
                Shot.name,
                AssetType.name,
                Asset.name,
            ).order_by(
                Episode.name,
                Sequence.name,
                Shot.name,
                AssetType.name,
                Asset.name,
            )
            for (
                entity_link,
                episode_name,
                sequence_name,
                shot_name,
                asset_type_name,
                asset_name,
            ) in query.all():
                results.append(
                    (
                        episode_name,
                        sequence_name,
                        "Shot",
                        shot_name,
                        asset_type_name,
                        asset_name,
                        entity_link.nb_occurences,
                        entity_link.label,
                    )
                )
        else:
            query = query.add_columns(
                Sequence.name,
                Shot.name,
                AssetType.name,
                Asset.name,
            ).order_by(
                Sequence.name,
                Shot.name,
                AssetType.name,
                Asset.name,
            )

            for (
                entity_link,
                sequence_name,
                shot_name,
                asset_type_name,
                asset_name,
            ) in query.all():
                results.append(
                    (
                        sequence_name,
                        "Shot",
                        shot_name,
                        asset_type_name,
                        asset_name,
                        entity_link.nb_occurences,
                        entity_link.label,
                    )
                )

        return results

    def build_asset_results(
        self, project_id, is_tv_show=False, episode_id=None
    ):
        results = []
        ParentAsset = aliased(Entity, name="parent_asset")
        ParentAssetType = aliased(EntityType, name="parent_asset_type")
        Asset = aliased(Entity, name="asset")
        AssetType = aliased(EntityType, name="asset_type")
        Episode = aliased(Entity, name="episode")
        shot_type = shots_service.get_shot_type()

        query = (
            EntityLink.query.join(
                ParentAsset, EntityLink.entity_in_id == ParentAsset.id
            )
            .join(
                ParentAssetType,
                ParentAsset.entity_type_id == ParentAssetType.id,
            )
            .join(Asset, EntityLink.entity_out_id == Asset.id)
            .join(AssetType, Asset.entity_type_id == AssetType.id)
            .filter(ParentAsset.project_id == project_id)
            .filter(ParentAssetType.id != shot_type["id"])
        )
        if is_tv_show:
            query = query.join(Episode, ParentAsset.source_id == Episode.id)
            if episode_id is not None:
                query = query.filter(Episode.id == episode_id)
            query = query.add_columns(
                Episode.name,
                ParentAssetType.name,
                ParentAsset.name,
                AssetType.name,
                Asset.name,
            ).order_by(
                Episode.name,
                ParentAssetType.name,
                ParentAsset.name,
                AssetType.name,
                Asset.name,
            )
            for (
                entity_link,
                episode_name,
                parent_asset_type_name,
                parent_name,
                asset_type_name,
                asset_name,
            ) in query.all():
                results.append(
                    (
                        episode_name,
                        "",
                        parent_asset_type_name,
                        parent_name,
                        asset_type_name,
                        asset_name,
                        entity_link.nb_occurences,
                        entity_link.label,
                    )
                )
            results += self.build_main_pack_results(project_id)
        else:
            query = query.add_columns(
                ParentAssetType.name,
                ParentAsset.name,
                AssetType.name,
                Asset.name,
            ).order_by(
                ParentAssetType.name,
                ParentAsset.name,
                AssetType.name,
                Asset.name,
            )
            for (
                entity_link,
                parent_asset_type_name,
                parent_name,
                asset_type_name,
                asset_name,
            ) in query.all():
                results.append(
                    (
                        parent_asset_type_name,
                        "",
                        parent_name,
                        asset_type_name,
                        asset_name,
                        entity_link.nb_occurences,
                        entity_link.label,
                    )
                )

        return results

    def build_main_pack_results(self, project_id):
        results = []
        ParentAsset = aliased(Entity, name="parent_asset")
        ParentAssetType = aliased(EntityType, name="parent_asset_type")
        Asset = aliased(Entity, name="asset")
        AssetType = aliased(EntityType, name="asset_type")
        shot_type = shots_service.get_shot_type()
        episode_type = shots_service.get_episode_type()

        query = (
            EntityLink.query.join(
                ParentAsset, EntityLink.entity_in_id == ParentAsset.id
            )
            .join(
                ParentAssetType,
                ParentAsset.entity_type_id == ParentAssetType.id,
            )
            .join(Asset, EntityLink.entity_out_id == Asset.id)
            .join(AssetType, Asset.entity_type_id == AssetType.id)
            .filter(ParentAsset.project_id == project_id)
            .filter(ParentAsset.source_id == None)
            .filter(ParentAssetType.id != shot_type["id"])
            .filter(ParentAssetType.id != episode_type["id"])
        )
        query = query.add_columns(
            ParentAssetType.name,
            ParentAsset.name,
            AssetType.name,
            Asset.name,
        ).order_by(
            ParentAssetType.name,
            ParentAsset.name,
            AssetType.name,
            Asset.name,
        )
        for (
            entity_link,
            parent_asset_type_name,
            parent_name,
            asset_type_name,
            asset_name,
        ) in query.all():
            results.append(
                (
                    "MP",
                    "",
                    parent_asset_type_name,
                    parent_name,
                    asset_type_name,
                    asset_name,
                    entity_link.nb_occurences,
                    entity_link.label,
                )
            )
        return results

    def build_episodes_results(self, project_id):
        results = []
        ParentAsset = aliased(Entity, name="parent_asset")
        ParentAssetType = aliased(EntityType, name="parent_asset_type")
        Asset = aliased(Entity, name="asset")
        AssetType = aliased(EntityType, name="asset_type")
        episode_type = shots_service.get_episode_type()

        query = (
            EntityLink.query.join(
                ParentAsset, EntityLink.entity_in_id == ParentAsset.id
            )
            .join(
                ParentAssetType,
                ParentAsset.entity_type_id == ParentAssetType.id,
            )
            .join(Asset, EntityLink.entity_out_id == Asset.id)
            .join(AssetType, Asset.entity_type_id == AssetType.id)
            .filter(ParentAsset.project_id == project_id)
            .filter(ParentAsset.source_id == None)
            .filter(ParentAssetType.id == episode_type["id"])
        )
        query = query.add_columns(
            ParentAssetType.name,
            ParentAsset.name,
            AssetType.name,
            Asset.name,
        ).order_by(
            ParentAssetType.name,
            ParentAsset.name,
            AssetType.name,
            Asset.name,
        )
        for (
            entity_link,
            parent_asset_type_name,
            parent_name,
            asset_type_name,
            asset_name,
        ) in query.all():
            results.append(
                (
                    "",
                    parent_asset_type_name,
                    parent_name,
                    asset_type_name,
                    asset_name,
                    entity_link.nb_occurences,
                    entity_link.label,
                )
            )
        return results



"""Add is hd_by_default column to organisation table

Revision ID: 1e2d77a2f0c4
Revises: 4e3738cdc34c
Create Date: 2021-10-12 18:56:43.683008

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "1e2d77a2f0c4"
down_revision = "4e3738cdc34c"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "organisation", sa.Column("hd_by_default", sa.Boolean(), nullable=True)
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("organisation", "hd_by_default")
    # ### end Alembic commands ###



"""add workflow to project

Revision ID: 5e2ce62632a6
Revises: c68c2a62cfac
Create Date: 2020-07-14 01:35:24.759248

"""

from alembic import op
import sqlalchemy as sa
import sqlalchemy_utils
import sqlalchemy_utils
import uuid

# revision identifiers, used by Alembic.
revision = "5e2ce62632a6"
down_revision = "c68c2a62cfac"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "project_asset_type_link",
        sa.Column(
            "project_id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.Column(
            "asset_type_id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["asset_type_id"],
            ["entity_type.id"],
        ),
        sa.ForeignKeyConstraint(
            ["project_id"],
            ["project.id"],
        ),
        sa.PrimaryKeyConstraint("project_id", "asset_type_id"),
    )
    op.create_table(
        "project_task_status_link",
        sa.Column(
            "project_id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.Column(
            "task_status_id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["project_id"],
            ["project.id"],
        ),
        sa.ForeignKeyConstraint(
            ["task_status_id"],
            ["task_status.id"],
        ),
        sa.PrimaryKeyConstraint("project_id", "task_status_id"),
    )
    op.create_table(
        "project_task_type_link",
        sa.Column(
            "project_id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.Column(
            "task_type_id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["project_id"],
            ["project.id"],
        ),
        sa.ForeignKeyConstraint(
            ["task_type_id"],
            ["task_type.id"],
        ),
        sa.PrimaryKeyConstraint("project_id", "task_type_id"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("project_task_type_link")
    op.drop_table("project_task_status_link")
    op.drop_table("project_asset_type_link")
    # ### end Alembic commands ###



"""
Utility routes to run full text search on Kitsu data.
"""

from flask import Blueprint
from zou.app.utils.api import configure_api_from_blueprint

from zou.app.blueprints.search.resources import SearchResource

routes = [("/data/search", SearchResource)]

blueprint = Blueprint("search", "search")
api = configure_api_from_blueprint(blueprint, routes)



import os
from zou.app.utils.string import strtobool


def envtobool(key, default=False):
    """
    Convert an environment variable to a boolean value.
    If environment variable can't be converted raise ValueError.
    """
    try:
        return strtobool(os.getenv(key, default))
    except ValueError:
        raise ValueError(
            f"Environment variable {key} cannot be converted to a boolean value."
        )


def env_with_semicolon_to_list(key, default=[]):
    """
    Convert an environment variable to a list.
    Items are separated by semicolon.
    """
    env_value = os.getenv(key, None)
    if env_value is None:
        return default
    else:
        return env_value.split(";")



"""add version field to software

Revision ID: d5665dca188b
Revises: 9683bd840dee
Create Date: 2025-07-14 12:44:29.378444

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "d5665dca188b"
down_revision = "9683bd840dee"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("software", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("version", sa.String(length=20), nullable=True)
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("software", schema=None) as batch_op:
        batch_op.drop_column("version")

    # ### end Alembic commands ###



"""Add TaskStatus.is_wip

Revision ID: 4d585b6956f2
Revises: d80f02824047
Create Date: 2025-06-30 01:11:29.146591

"""

from alembic import op
import sqlalchemy as sa
from alembic import op
import sqlalchemy as sa
from sqlalchemy.orm.session import Session
from sqlalchemy.ext.declarative import declarative_base
from zou.migrations.utils.base import BaseMixin
from sqlalchemy.sql import expression


# revision identifiers, used by Alembic.
revision = "4d585b6956f2"
down_revision = "d80f02824047"
branch_labels = None
depends_on = None

base = declarative_base()


class TaskStatus(base, BaseMixin):
    """
    Describe the state of a task. A status marked as reviewable expects a
    preview file linked to relate comment.
    """

    __tablename__ = "task_status"
    name = sa.Column(sa.String(40), nullable=False)
    archived = sa.Column(sa.Boolean(), default=False)
    short_name = sa.Column(
        sa.String(10), unique=True, nullable=False, index=True
    )
    description = sa.Column(sa.Text())
    color = sa.Column(sa.String(7), nullable=False)
    priority = sa.Column(sa.Integer, default=1)

    is_done = sa.Column(sa.Boolean(), default=False, index=True)
    is_artist_allowed = sa.Column(sa.Boolean(), default=True)
    is_client_allowed = sa.Column(sa.Boolean(), default=True)
    is_retake = sa.Column(sa.Boolean(), default=False)
    is_feedback_request = sa.Column(sa.Boolean(), default=False, index=True)
    is_default = sa.Column(sa.Boolean(), default=False, index=True)
    is_wip = sa.Column(
        sa.Boolean(),
        server_default=expression.false(),
        default=False,
        index=True,
    )
    shotgun_id = sa.Column(sa.Integer)

    for_concept = sa.Column(
        sa.Boolean(), server_default=expression.false(), default=False
    )


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("task_status", schema=None) as batch_op:
        batch_op.add_column(sa.Column("is_wip", sa.Boolean(), nullable=True))
        batch_op.create_index(
            batch_op.f("ix_task_status_is_wip"), ["is_wip"], unique=False
        )
    session = Session(bind=op.get_bind())
    session.query(TaskStatus).update(
        {
            TaskStatus.is_wip: TaskStatus.short_name == "wip",
        }
    )
    session.commit()

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("task_status", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_task_status_is_wip"))
        batch_op.drop_column("is_wip")

    # ### end Alembic commands ###



import csv

from io import StringIO
from flask import make_response
from slugify import slugify


def build_csv_response(csv_content, file_name="export"):
    """
    Construct a Flask response that returns content of a csv a file.
    """
    file_name = build_csv_file_name(file_name)
    response_body = build_csv_string(csv_content)
    csv_response = make_response(response_body)
    csv_response = build_csv_headers(csv_response, file_name)

    return csv_response


def build_csv_file_name(file_name):
    """
    Add application name as prefix of the file name.
    """
    return "kitsu_%s" % slugify(file_name, separator="_")


def build_csv_string(csv_content):
    """
    Build a CSV formatted string from an array.
    """
    string_wrapper = StringIO()
    csv_writer = csv.writer(string_wrapper, delimiter=";")
    csv_writer.writerows(csv_content)
    return string_wrapper.getvalue()


def build_csv_headers(csv_response, file_name):
    """
    Build HTTP response headers needed to return CSV content as a file.
    """
    csv_response.headers["Content-Disposition"] = (
        "attachment; filename=%s.csv" % file_name
    )
    csv_response.headers["Content-type"] = "text/csv"
    return csv_response



from datetime import timedelta
from operator import itemgetter
from sqlalchemy.orm import aliased
from sqlalchemy.exc import IntegrityError, StatementError
from sqlalchemy import func, or_

from zou.app.utils import (
    cache,
    date_helpers,
    events,
    fields,
    query as query_utils,
)

from zou.app.models.entity import (
    Entity,
    EntityLink,
    EntityVersion,
    EntityConceptLink,
)
from zou.app.models.person import Person
from zou.app.models.project import Project
from zou.app.models.preview_file import PreviewFile
from zou.app.models.schedule_item import ScheduleItem
from zou.app.models.subscription import Subscription
from zou.app.models.task import Task, TaskPersonLink
from zou.app.models.time_spent import TimeSpent

from zou.app.services import (
    deletion_service,
    entities_service,
    persons_service,
    projects_service,
    notifications_service,
    names_service,
    user_service,
    index_service,
    concepts_service,
)
from zou.app.services.exception import (
    EpisodeNotFoundException,
    ModelWithRelationsDeletionException,
    SequenceNotFoundException,
    SceneNotFoundException,
    ShotNotFoundException,
    WrongIdFormatException,
)


def clear_shot_cache(shot_id):
    cache.cache.delete_memoized(get_shot, shot_id)
    cache.cache.delete_memoized(get_shot, shot_id, True)
    cache.cache.delete_memoized(get_full_shot, shot_id)


def clear_sequence_cache(sequence_id):
    cache.cache.delete_memoized(get_sequence, sequence_id)
    cache.cache.delete_memoized(get_full_sequence, sequence_id)


def clear_episode_cache(episode_id):
    cache.cache.delete_memoized(get_episode, episode_id)
    cache.cache.delete_memoized(get_episode_by_name)


@cache.memoize_function(1200)
def get_episode_type():
    return entities_service.get_temporal_entity_type_by_name("Episode")


@cache.memoize_function(1200)
def get_edit_type():
    return entities_service.get_temporal_entity_type_by_name("Edit")


@cache.memoize_function(1200)
def get_sequence_type():
    return entities_service.get_temporal_entity_type_by_name("Sequence")


@cache.memoize_function(1200)
def get_shot_type():
    return entities_service.get_temporal_entity_type_by_name("Shot")


@cache.memoize_function(1200)
def get_scene_type():
    return entities_service.get_temporal_entity_type_by_name("Scene")


@cache.memoize_function(1200)
def get_camera_type():
    return entities_service.get_entity_type_by_name("Camera")


def get_episodes(criterions={}):
    """
    Get all episodes for given criterions.
    """
    episode_type = get_episode_type()
    criterions["entity_type_id"] = episode_type["id"]
    query = Entity.query.order_by(Entity.name)
    query = query_utils.apply_criterions_to_db_query(Entity, query, criterions)
    try:
        episodes = query.all()
    except StatementError:  # Occurs when an id is not properly formatted
        raise WrongIdFormatException
    return Entity.serialize_list(episodes, obj_type="Episode")


def get_sequences(criterions={}):
    """
    Get all sequences for given criterions.
    """
    sequence_type = get_sequence_type()
    criterions["entity_type_id"] = sequence_type["id"]
    query = Entity.query.order_by(Entity.name)
    query = query_utils.apply_criterions_to_db_query(Entity, query, criterions)
    try:
        sequences = query.all()
    except StatementError:  # Occurs when an id is not properly formatted
        raise WrongIdFormatException
    return Entity.serialize_list(sequences, obj_type="Sequence")


def get_shots(criterions={}):
    """
    Get all shots for given criterions.
    """
    shot_type = get_shot_type()
    criterions["entity_type_id"] = shot_type["id"]
    Sequence = aliased(Entity, name="sequence")
    is_only_assignation = "assigned_to" in criterions
    if is_only_assignation:
        del criterions["assigned_to"]

    query = Entity.query
    query = query_utils.apply_criterions_to_db_query(Entity, query, criterions)
    query = (
        query.join(Project, Project.id == Entity.project_id)
        .join(Sequence, Sequence.id == Entity.parent_id)
        .add_columns(Project.name)
        .add_columns(Sequence.name)
        .order_by(Entity.name)
    )

    if is_only_assignation:
        query = query.outerjoin(Task, Task.entity_id == Entity.id)
        query = query.filter(user_service.build_assignee_filter())

    try:
        data = query.all()
    except StatementError:  # Occurs when an id is not properly formatted
        raise WrongIdFormatException

    shots = []
    for shot_model, project_name, sequence_name in data:
        shot = shot_model.serialize(obj_type="Shot")
        shot["project_name"] = project_name
        shot["sequence_name"] = sequence_name
        shots.append(shot)

    return shots


def get_scenes(criterions={}):
    """
    Get all scenes for given criterions.
    """
    scene_type = get_scene_type()
    criterions["entity_type_id"] = scene_type["id"]
    Sequence = aliased(Entity, name="sequence")

    is_only_assignation = "assigned_to" in criterions
    if is_only_assignation:
        del criterions["assigned_to"]

    query = Entity.query
    query = query_utils.apply_criterions_to_db_query(Entity, query, criterions)
    query = (
        query.join(Project, Entity.project_id == Project.id)
        .join(Sequence, Sequence.id == Entity.parent_id)
        .add_columns(Project.name)
        .add_columns(Sequence.name)
    )

    if is_only_assignation:
        query = query.outerjoin(Task, Task.entity_id == Entity.id)
        query = query.filter(user_service.build_assignee_filter())

    try:
        data = query.all()
    except StatementError:  # Occurs when an id is not properly formatted
        raise WrongIdFormatException

    scenes = []
    for scene_model, project_name, sequence_name in data:
        scene = scene_model.serialize(obj_type="Scene")
        scene["project_name"] = project_name
        scene["sequence_name"] = sequence_name
        scenes.append(scene)

    return scenes


def get_episode_map(criterions={}):
    """
    Returns a dict where keys are episode_id and values are episodes.
    """
    episodes = get_episodes(criterions)
    episode_map = {}
    for episode in episodes:
        episode_map[episode["id"]] = episode
    return episode_map


def get_shots_and_tasks(criterions={}):
    """
    Get all shots for given criterions with related tasks for each shot.
    """
    shot_type = get_shot_type()
    shot_map = {}
    task_map = {}
    subscription_map = notifications_service.get_subscriptions_for_user(
        criterions.get("project_id", None), get_shot_type()["id"]
    )

    Sequence = aliased(Entity, name="sequence")
    Episode = aliased(Entity, name="episode")

    query = (
        Entity.query.join(Project, Project.id == Entity.project_id)
        .join(Sequence, Sequence.id == Entity.parent_id)
        .outerjoin(Episode, Episode.id == Sequence.parent_id)
        .outerjoin(Task, Task.entity_id == Entity.id)
        .outerjoin(TaskPersonLink)
        .add_columns(
            Episode.name,
            Episode.id,
            Sequence.name,
            Sequence.id,
            Task.id,
            Task.task_type_id,
            Task.task_status_id,
            Task.priority,
            Task.estimation,
            Task.duration,
            Task.retake_count,
            Task.real_start_date,
            Task.end_date,
            Task.start_date,
            Task.due_date,
            Task.done_date,
            Task.last_comment_date,
            Task.last_preview_file_id,
            Task.nb_assets_ready,
            Task.difficulty,
            Task.nb_drawings,
            TaskPersonLink.person_id,
            Project.id,
            Project.name,
        )
        .filter(Entity.entity_type_id == shot_type["id"])
    )
    if "id" in criterions:
        query = query.filter(Entity.id == criterions["id"])

    if "project_id" in criterions:
        query = query.filter(Entity.project_id == criterions["project_id"])

    if "episode_id" in criterions and criterions["episode_id"] != "all":
        query = query.filter(Sequence.parent_id == criterions["episode_id"])

    if "assigned_to" in criterions:
        query = query.filter(user_service.build_assignee_filter())
        del criterions["assigned_to"]

    query_result = query.all()

    if "vendor_departments" in criterions:
        not_allowed_descriptors_field_names = (
            entities_service.get_not_allowed_descriptors_fields_for_vendor(
                "Shot",
                criterions["vendor_departments"],
                set(shot[0].project_id for shot in query_result),
            )
        )

    for (
        shot,
        episode_name,
        episode_id,
        sequence_name,
        sequence_id,
        task_id,
        task_type_id,
        task_status_id,
        task_priority,
        task_estimation,
        task_duration,
        task_retake_count,
        task_real_start_date,
        task_end_date,
        task_start_date,
        task_due_date,
        task_done_date,
        task_last_comment_date,
        task_last_preview_file_id,
        task_nb_assets_ready,
        task_difficulty,
        task_nb_drawings,
        person_id,
        project_id,
        project_name,
    ) in query_result:
        shot_id = str(shot.id)

        if shot_id not in shot_map:
            data = fields.serialize_value(shot.data or {})
            if "vendor_departments" in criterions:
                data = (
                    entities_service.remove_not_allowed_fields_from_metadata(
                        not_allowed_descriptors_field_names[shot.project_id],
                        data,
                    )
                )

            shot_map[shot_id] = fields.serialize_dict(
                {
                    "canceled": shot.canceled,
                    "data": data,
                    "description": shot.description,
                    "entity_type_id": shot.entity_type_id,
                    "episode_id": episode_id,
                    "episode_name": episode_name or "",
                    "fps": data.get("fps", None),
                    "frame_in": data.get("frame_in", None),
                    "frame_out": data.get("frame_out", None),
                    "id": shot.id,
                    "name": shot.name,
                    "nb_frames": shot.nb_frames,
                    "parent_id": shot.parent_id,
                    "preview_file_id": shot.preview_file_id or None,
                    "project_id": project_id,
                    "project_name": project_name,
                    "sequence_id": sequence_id,
                    "sequence_name": sequence_name,
                    "source_id": shot.source_id,
                    "nb_entities_out": shot.nb_entities_out,
                    "is_casting_standby": shot.is_casting_standby,
                    "tasks": [],
                    "type": "Shot",
                }
            )

        if task_id is not None:
            task_id = str(task_id)
            if task_id not in task_map:
                task_dict = fields.serialize_dict(
                    {
                        "id": task_id,
                        "duration": task_duration,
                        "due_date": task_due_date,
                        "end_date": task_end_date,
                        "done_date": task_done_date,
                        "entity_id": shot_id,
                        "estimation": task_estimation,
                        "is_subscribed": subscription_map.get(task_id, False),
                        "last_comment_date": task_last_comment_date,
                        "last_preview_file_id": task_last_preview_file_id,
                        "nb_assets_ready": task_nb_assets_ready,
                        "priority": task_priority or 0,
                        "real_start_date": task_real_start_date,
                        "retake_count": task_retake_count,
                        "start_date": task_start_date,
                        "difficulty": task_difficulty,
                        "nb_drawings": task_nb_drawings,
                        "task_status_id": task_status_id,
                        "task_type_id": task_type_id,
                        "assignees": [],
                    }
                )
                task_map[task_id] = task_dict
                shot_dict = shot_map[shot_id]
                shot_dict["tasks"].append(task_dict)

            if person_id:
                task_map[task_id]["assignees"].append(str(person_id))

    return list(shot_map.values())


def get_shot_raw(shot_id):
    """
    Return given shot as an active record.
    """
    shot_type = get_shot_type()
    try:
        shot = Entity.get_by(entity_type_id=shot_type["id"], id=shot_id)
    except StatementError:
        raise ShotNotFoundException

    if shot is None:
        raise ShotNotFoundException

    return shot


@cache.memoize_function(120)
def get_shot(shot_id, relations=False):
    """
    Return given shot as a dictionary.
    """
    return get_shot_raw(shot_id).serialize(
        obj_type="Shot", relations=relations
    )


@cache.memoize_function(120)
def get_full_shot(shot_id):
    """
    Return given shot as a dictionary with extra data like project and
    sequence names.
    """
    shots = get_shots_and_tasks({"id": shot_id})
    if len(shots) > 0:
        shot = shots[0]
        shot.update(get_shot(shot_id, relations=True))
        return shot
    else:
        raise ShotNotFoundException


def get_scene_raw(scene_id):
    """
    Return given scene as an active record.
    """
    scene_type = get_scene_type()
    try:
        scene = Entity.get_by(entity_type_id=scene_type["id"], id=scene_id)
    except StatementError:
        raise SceneNotFoundException

    if scene is None:
        raise SceneNotFoundException

    return scene


def get_scene(scene_id):
    """
    Return given scene as a dictionary.
    """
    return get_scene_raw(scene_id).serialize(obj_type="Scene")


def get_full_scene(scene_id):
    """
    Return given scene as a dictionary with extra data like project and sequence
    names.
    """
    scene = get_scene(scene_id)
    project = projects_service.get_project(scene["project_id"])
    sequence = get_sequence(scene["parent_id"])
    scene["project_name"] = project["name"]
    scene["sequence_id"] = sequence["id"]
    scene["sequence_name"] = sequence["name"]
    if sequence["parent_id"] is not None:
        episode = get_episode(sequence["parent_id"])
        scene["episode_id"] = episode["id"]
        scene["episode_name"] = episode["name"]

    return scene


def get_sequence_raw(sequence_id):
    """
    Return given sequence as an active record.
    """
    sequence_type = get_sequence_type()
    try:
        sequence = Entity.get_by(
            entity_type_id=sequence_type["id"], id=sequence_id
        )
    except StatementError:
        raise SequenceNotFoundException

    if sequence is None:
        raise SequenceNotFoundException

    return sequence


@cache.memoize_function(120)
def get_sequence(sequence_id):
    """
    Return given sequence as a dictionary.
    """
    return get_sequence_raw(sequence_id).serialize(obj_type="Sequence")


@cache.memoize_function(120)
def get_full_sequence(sequence_id):
    """
    Return given sequence as a dictionary with extra data like project name.
    """
    sequence = get_sequence(sequence_id)
    project = projects_service.get_project(sequence["project_id"])
    sequence["project_name"] = project["name"]

    if sequence["parent_id"] is not None:
        episode = get_episode(sequence["parent_id"])
        sequence["episode_id"] = episode["id"]
        sequence["episode_name"] = episode["name"]

    return sequence


def get_sequence_from_shot(shot):
    """
    Return parent sequence of given shot.
    """
    try:
        sequence = Entity.get(shot["parent_id"])
    except Exception:
        raise SequenceNotFoundException("Wrong parent_id for given shot.")
    return sequence.serialize(obj_type="Sequence")


def get_episode_raw(episode_id):
    """
    Return given episode as an active record.
    """
    episode_type = get_episode_type()
    if episode_type is None:
        episode_type = get_episode_type()

    try:
        episode = Entity.get_by(
            entity_type_id=episode_type["id"], id=episode_id
        )
    except StatementError:
        raise EpisodeNotFoundException

    if episode is None:
        raise EpisodeNotFoundException
    return episode


@cache.memoize_function(120)
def get_episode(episode_id):
    """
    Return given episode as a dictionary.
    """
    return get_episode_raw(episode_id).serialize(obj_type="Episode")


@cache.memoize_function(120)
def get_episode_by_name(project_id, episode_name):
    """
    Get episode matching given name project_id. Raises an exception if episode
    is not found.
    """
    episode_type_id = get_episode_type()["id"]
    episode = (
        Entity.query.filter(Entity.entity_type_id == episode_type_id)
        .filter(Entity.project_id == project_id)
        .filter(Entity.name.ilike(episode_name))
        .first()
    )

    if episode is None:
        raise EpisodeNotFoundException()

    return episode.serialize(obj_type="Episode")


def get_full_episode(episode_id):
    """
    Return given episode as a dictionary with extra data like project name.
    """
    episode = get_episode(episode_id)
    project = projects_service.get_project(episode["project_id"])
    episode["project_name"] = project["name"]
    return episode


def get_episode_from_sequence(sequence):
    """
    Return parent episode of given sequence.
    """
    try:
        episode = Entity.get(sequence["parent_id"])
    except Exception:
        raise EpisodeNotFoundException("Wrong parent_id for given sequence.")
    return episode.serialize(obj_type="Episode")


def get_shot_by_shotgun_id(shotgun_id):
    """
    Retrieves a shot identifed by its shotgun ID (stored during import).
    """
    shot_type = get_shot_type()
    shot = Entity.get_by(entity_type_id=shot_type["id"], shotgun_id=shotgun_id)
    if shot is None:
        raise ShotNotFoundException

    return shot.serialize(obj_type="Shot")


def get_scene_by_shotgun_id(shotgun_id):
    """
    Retrieves a scene identifed by its shotgun ID (stored during import).
    """
    scene_type = get_scene_type()
    scene = Entity.get_by(
        entity_type_id=scene_type["id"], shotgun_id=shotgun_id
    )
    if scene is None:
        raise SceneNotFoundException

    return scene.serialize(obj_type="Scene")


def get_sequence_by_shotgun_id(shotgun_id):
    """
    Retrieves a sequence identifed by its shotgun ID (stored during import).
    """
    sequence_type = get_sequence_type()
    sequence = Entity.get_by(
        entity_type_id=sequence_type["id"], shotgun_id=shotgun_id
    )
    if sequence is None:
        raise SequenceNotFoundException

    return sequence.serialize(obj_type="Sequence")


def get_episode_by_shotgun_id(shotgun_id):
    """
    Retrieves an episode identifed by its shotgun ID (stored during import).
    """
    episode_type = get_episode_type()
    episode = Entity.get_by(
        entity_type_id=episode_type["id"], shotgun_id=shotgun_id
    )
    if episode is None:
        raise EpisodeNotFoundException

    return episode.serialize(obj_type="Episode")


def is_shot(entity):
    """
    Returns True if given entity has 'Shot' as entity type
    """
    shot_type = get_shot_type()
    return str(entity["entity_type_id"]) == shot_type["id"]


def is_scene(entity):
    """
    Returns True if given entity has 'Scene' as entity type
    """
    scene_type = get_scene_type()
    return str(entity["entity_type_id"]) == scene_type["id"]


def is_sequence(entity):
    """
    Returns True if given entity has 'Sequence' as entity type
    """
    sequence_type = get_sequence_type()
    return str(entity["entity_type_id"]) == sequence_type["id"]


def is_edit(entity):
    """
    Returns True if given entity has 'Edit' as entity type
    """
    edit_type = get_edit_type()
    return str(entity["entity_type_id"]) == edit_type["id"]


def is_episode(entity):
    """
    Returns True if given entity has 'Episode' as entity type
    """
    episode_type = get_episode_type()
    return str(entity["entity_type_id"]) == episode_type["id"]


def get_or_create_first_episode(project_id, created_by=None):
    """
    Get the first episode of the production.
    """
    episode = (
        Entity.query.filter_by(project_id=project_id)
        .order_by(Entity.name)
        .first()
    )
    if episode is not None:
        return episode.serialize()
    else:
        return create_episode(
            project_id, "running", "E01", created_by=created_by
        )


def get_episodes_for_project(project_id, only_assigned=False):
    """
    Retrieve all episodes related to given project.
    """
    if only_assigned:
        Sequence = aliased(Entity, name="sequence")
        Shot = aliased(Entity, name="shot")
        Asset = aliased(Entity, name="asset")
        query = (
            Entity.query.join(Sequence, Entity.id == Sequence.parent_id)
            .join(Shot, Sequence.id == Shot.parent_id)
            .join(Task, Shot.id == Task.entity_id)
            .filter(Entity.project_id == project_id)
            .filter(user_service.build_assignee_filter())
        )
        shot_episodes = fields.serialize_models(query.all())
        shot_episode_ids = {episode["id"]: True for episode in shot_episodes}
        query = (
            Entity.query.join(Asset, Entity.id == Asset.source_id)
            .join(Task, Asset.id == Task.entity_id)
            .filter(Entity.project_id == project_id)
            .filter(user_service.build_assignee_filter())
        )
        asset_episodes = fields.serialize_models(query.all())
        result = shot_episodes
        for episode in asset_episodes:
            if episode["id"] not in shot_episode_ids:
                result.append(episode)
        return result
    else:
        return entities_service.get_entities_for_project(
            project_id, get_episode_type()["id"], "Episode"
        )


def get_sequences_for_project(project_id, only_assigned=False):
    """
    Retrieve all sequences related to given project.
    """
    if only_assigned:
        Shot = aliased(Entity, name="shot")
        query = (
            Entity.query.join(Shot, Entity.id == Shot.parent_id)
            .join(Task, Shot.id == Task.entity_id)
            .filter(Entity.project_id == project_id)
            .filter(user_service.build_assignee_filter())
        )
        return fields.serialize_models(query.all())
    else:
        return entities_service.get_entities_for_project(
            project_id, get_sequence_type()["id"], "Sequence"
        )


def get_sequences_for_episode(episode_id, only_assigned=False):
    """
    Retrieve all sequences related to given episode.
    """
    if only_assigned:
        Shot = aliased(Entity, name="shot")
        query = (
            Entity.query.join(Shot, Entity.id == Shot.parent_id)
            .join(Task, Shot.id == Task.entity_id)
            .filter(Entity.parent_id == episode_id)
            .filter(user_service.build_assignee_filter())
        )
        return fields.serialize_models(query.all())
    else:
        return get_sequences({"parent_id": episode_id})


def get_shots_for_project(project_id, only_assigned=False):
    """
    Retrieve all shots related to given project.
    """
    return entities_service.get_entities_for_project(
        project_id, get_shot_type()["id"], "Shot", only_assigned=only_assigned
    )


def get_shots_for_episode(episode_id, relations=False):
    """
    Get all shots for given episode.
    """
    Sequence = aliased(Entity, name="sequence")
    shot_type_id = get_shot_type()["id"]
    result = (
        Entity.query.filter(Entity.entity_type_id == shot_type_id)
        .filter(Sequence.parent_id == episode_id)
        .join(Sequence, Entity.parent_id == Sequence.id)
    ).all()
    return Entity.serialize_list(result, "Shot", relations=relations)


def get_scenes_for_project(project_id, only_assigned=False):
    """
    Retrieve all scenes related to given project.
    """
    return entities_service.get_entities_for_project(
        project_id,
        get_scene_type()["id"],
        "Scene",
        only_assigned=only_assigned,
    )


def get_scenes_for_sequence(sequence_id):
    """
    Retrieve all scenes children of given sequence.
    """
    get_sequence(sequence_id)
    scene_type_id = get_scene_type()["id"]
    result = (
        Entity.query.filter(Entity.entity_type_id == scene_type_id)
        .filter(Entity.parent_id == sequence_id)
        .order_by(Entity.name)
        .all()
    )
    return Entity.serialize_list(result, "Scene")


def remove_shot(shot_id, force=False):
    """
    Remove given shot from database. If it has tasks linked to it, it marks
    the shot as canceled. Deletion can be forced.
    """
    shot = get_shot_raw(shot_id)
    is_tasks_related = Task.query.filter_by(entity_id=shot_id).count() > 0

    if is_tasks_related and not force:
        shot.update({"canceled": True})
        clear_shot_cache(shot_id)
        events.emit(
            "shot:update",
            {"shot_id": shot_id},
            project_id=str(shot.project_id),
        )
    else:
        from zou.app.services import tasks_service

        tasks = Task.query.filter_by(entity_id=shot_id).all()
        for task in tasks:
            deletion_service.remove_task(task.id, force=True)
            tasks_service.clear_task_cache(str(task.id))

        EntityVersion.delete_all_by(entity_id=shot_id)
        Subscription.delete_all_by(entity_id=shot_id)
        EntityLink.delete_all_by(entity_in_id=shot_id)
        EntityLink.delete_all_by(entity_out_id=shot_id)
        EntityConceptLink.delete_all_by(entity_in_id=shot_id)
        EntityConceptLink.delete_all_by(entity_out_id=shot_id)

        shot.delete()
        events.emit(
            "shot:delete",
            {"shot_id": shot_id},
            project_id=str(shot.project_id),
        )
        index_service.remove_shot_index(shot_id)
        clear_shot_cache(shot_id)

    deleted_shot = shot.serialize(obj_type="Shot")
    return deleted_shot


def remove_scene(scene_id):
    """
    Remove given scene from database. If it has tasks linked to it, it marks
    the scene as canceled.
    """
    scene = get_scene_raw(scene_id)
    try:
        scene.delete()
    except IntegrityError:
        scene.update({"canceled": True})
    deleted_scene = scene.serialize(obj_type="Scene")
    events.emit(
        "scene:delete",
        {"scene_id": scene_id},
        project_id=str(scene.project_id),
    )
    return deleted_scene


def remove_sequence(sequence_id, force=False):
    """
    Remove a sequence and all related shots.
    """
    sequence = get_sequence_raw(sequence_id)
    if force:
        from zou.app.services import tasks_service

        for shot in Entity.get_all_by(parent_id=sequence_id):
            remove_shot(shot.id, force=True)
        Subscription.delete_all_by(entity_id=sequence_id)
        ScheduleItem.delete_all_by(object_id=sequence_id)

        tasks = Task.query.filter_by(entity_id=sequence_id).all()
        for task in tasks:
            deletion_service.remove_task(task.id, force=True)
            tasks_service.clear_task_cache(str(task.id))
        Subscription.delete_all_by(entity_id=sequence_id)
    try:
        sequence.delete()
        events.emit(
            "sequence:delete",
            {"sequence_id": sequence_id},
            project_id=str(sequence.project_id),
        )
    except IntegrityError:
        raise ModelWithRelationsDeletionException(
            "Some data are still linked to this sequence."
        )
    clear_sequence_cache(sequence_id)
    return sequence.serialize(obj_type="Sequence")


def create_episode(
    project_id,
    name,
    status="running",
    description="",
    data={},
    created_by=None,
):
    """
    Create episode for given project.
    """
    episode_type = get_episode_type()
    episode = Entity.get_by(
        entity_type_id=episode_type["id"], project_id=project_id, name=name
    )
    if status not in ["running", "complete", "standby", "canceled"]:
        status = "running"
    if episode is None:
        episode = Entity.create(
            entity_type_id=episode_type["id"],
            project_id=project_id,
            name=name,
            status=status,
            description=description,
            data=data,
            created_by=created_by,
        )
        events.emit(
            "episode:new", {"episode_id": episode.id}, project_id=project_id
        )
    return episode.serialize(obj_type="Episode")


def create_sequence(
    project_id, episode_id, name, description="", data={}, created_by=None
):
    """
    Create sequence for given project and episode.
    """
    sequence_type = get_sequence_type()

    if episode_id is not None:
        get_episode(episode_id)  # raises EpisodeNotFound if it fails.

    sequence = Entity.get_by(
        entity_type_id=sequence_type["id"],
        parent_id=episode_id,
        project_id=project_id,
        name=name,
    )
    if sequence is None:
        sequence = Entity.create(
            entity_type_id=sequence_type["id"],
            project_id=project_id,
            parent_id=episode_id,
            name=name,
            description=description,
            data=data,
            created_by=created_by,
        )
        events.emit(
            "sequence:new", {"sequence_id": sequence.id}, project_id=project_id
        )
    return sequence.serialize(obj_type="Sequence")


def create_shot(
    project_id,
    sequence_id,
    name,
    data={},
    nb_frames=0,
    description=None,
    created_by=None,
):
    """
    Create shot for given project and sequence.
    """
    shot_type = get_shot_type()

    if sequence_id is not None:
        # raises SequenceNotFound if it fails.
        sequence = get_sequence(sequence_id)

    shot = Entity.get_by(
        entity_type_id=shot_type["id"],
        parent_id=sequence_id,
        project_id=project_id,
        name=name,
    )
    if shot is None:
        shot = Entity.create(
            entity_type_id=shot_type["id"],
            project_id=project_id,
            parent_id=sequence_id,
            name=name,
            data=data,
            nb_frames=nb_frames,
            description=description,
            created_by=created_by,
        )

        index_service.index_shot(shot)
        events.emit(
            "shot:new",
            {
                "shot_id": shot.id,
                "episode_id": sequence["parent_id"],
            },
            project_id=project_id,
        )

    return shot.serialize(obj_type="Shot")


def create_scene(project_id, sequence_id, name, created_by=None):
    """
    Create scene for given project and sequence.
    """
    scene_type = get_scene_type()

    if sequence_id is not None:
        # raises SequenceNotFound if it fails.
        sequence = get_sequence(sequence_id)
        if sequence["project_id"] != project_id:
            raise SequenceNotFoundException

    scene = Entity.get_by(
        entity_type_id=scene_type["id"],
        parent_id=sequence_id,
        project_id=project_id,
        name=name,
    )
    if scene is None:
        scene = Entity.create(
            entity_type_id=scene_type["id"],
            project_id=project_id,
            parent_id=sequence_id,
            name=name,
            data={},
            created_by=created_by,
        )
        events.emit("scene:new", {"scene_id": scene.id}, project_id=project_id)
    return scene.serialize(obj_type="Scene")


def update_shot(shot_id, data_dict):
    """
    Update shot fields matching given id with data from dict given in parameter.
    """
    shot = get_shot_raw(shot_id)
    shot.update(data_dict)

    index_service.remove_shot_index(shot.id)
    index_service.index_shot(shot)
    clear_shot_cache(shot_id)
    events.emit(
        "shot:update", {"shot_id": shot_id}, project_id=str(shot.project_id)
    )

    return shot.serialize()


def get_shot_versions(shot_id):
    """
    Shot metadata changes are versioned. This function returns all versions
    of a given shot.
    """
    versions = (
        EntityVersion.query.filter_by(entity_id=shot_id)
        .order_by(EntityVersion.created_at.desc())
        .all()
    )
    return EntityVersion.serialize_list(versions, obj_type="ShotVersion")


def get_base_entity_type_name(entity_dict):
    type_name = "Asset"
    if is_shot(entity_dict):
        type_name = "Shot"
    elif is_sequence(entity_dict):
        type_name = "Sequence"
    elif is_edit(entity_dict):
        type_name = "Edit"
    elif is_episode(entity_dict):
        type_name = "Episode"
    elif is_scene(entity_dict):
        type_name = "Scene"
    elif concepts_service.is_concept(entity_dict):
        type_name = "Concept"

    return type_name


def get_weighted_quotas(
    project_id,
    task_type_id=None,
    person_id=None,
    studio_id=None,
    feedback=True,
):
    """
    Build quota statistics. It counts the number of frames done for each day.
    A shot is considered done at the first feedback request or at last
    approval.

    If time spent is  filled for it, it weights the result with the frame
    number with the time spents. If there is no time spent, it considers that
    the work was done from the wip date to the feedback date (or approval date).
    It computes the shot count and the number of seconds too.

    If the `feedback` flag is set to True, it uses the feedback date
    (real_end_date), if feedback is set to False, it uses the approval date
    (done_date).
    """
    fps = projects_service.get_project_fps(project_id)
    timezone = user_service.get_timezone()
    shot_type = get_shot_type()
    quotas = {}
    query = (
        Task.query.filter(Entity.entity_type_id == shot_type["id"])
        .filter(Task.project_id == project_id)
        .join(Entity, Entity.id == Task.entity_id)
        .join(Project, Project.id == Task.project_id)
        .join(TimeSpent, Task.id == TimeSpent.task_id)
        .add_columns(
            Entity.nb_frames,
            TimeSpent.date,
            TimeSpent.duration,
            TimeSpent.person_id,
        )
    )

    if task_type_id is not None:
        query = query.filter(Task.task_type_id == task_type_id)

    if person_id is not None:
        query = query.filter(TimeSpent.person_id == person_id)

    if feedback:
        query = query.filter(Task.end_date != None)
    else:
        query = query.filter(Task.done_date != None)

    if studio_id is not None:
        persons_from_studio = Person.query.filter(
            Person.studio_id == studio_id
        ).all()
        query = query.filter(
            or_(*[Task.assignees.contains(p) for p in persons_from_studio])
        )
    result = query.all()

    for task, nb_frames, date, duration, task_person_id in result:
        task_person_id = str(task_person_id)
        nb_drawings = task.nb_drawings or 0
        nb_frames = nb_frames or 0
        if task.duration > 0:
            nb_frames = round(nb_frames * (duration / task.duration))
            nb_drawings = round(nb_drawings * (duration / task.duration))
            entry_id = str(task_person_id)
            # We get quotas for a specific person split by task types
            if person_id is not None:
                entry_id = str(task.task_type_id)
            for entry in [entry_id, "total"]:
                _add_quota_entry(
                    quotas, entry, date, timezone, nb_frames, nb_drawings, fps
                )

    query = (
        Task.query.filter(Task.project_id == project_id)
        .filter(Entity.entity_type_id == shot_type["id"])
        .filter(Task.task_type_id == task_type_id)
        .filter(Task.real_start_date != None)
        .filter(TimeSpent.id == None)
        .join(Entity, Entity.id == Task.entity_id)
        .join(Project, Project.id == Task.project_id)
        .outerjoin(TimeSpent, Task.id == TimeSpent.task_id)
        .join(Task.assignees)
        .add_columns(Entity.nb_frames, Person.id)
    )

    if task_type_id is not None:
        query = query.filter(Task.task_type_id == task_type_id)

    if person_id is not None:
        person = persons_service.get_person_raw(person_id)
        query = query.filter(Task.assignees.contains(person))

    if feedback:
        query = query.filter(Task.end_date != None)
    else:
        query = query.filter(Task.done_date != None)

    if studio_id is not None:
        query = query.filter(
            or_(*[Task.assignees.contains(p) for p in persons_from_studio])
        )
    result = query.all()

    for task, nb_frames, task_person_id in result:
        date = task.done_date
        if feedback:
            date = task.end_date

        business_days = (
            date_helpers.get_business_days(task.real_start_date, date) + 1
        )
        if nb_frames is not None:
            nb_frames = round(nb_frames / business_days) or 0
        else:
            nb_frames = 0

        nb_drawings = task.nb_drawings or 0

        for x in range((date - task.real_start_date).days + 1):
            if date.weekday() < 5:
                entry_id = str(task_person_id)
                # We get quotas for a specific person split by task types
                if person_id is not None:
                    entry_id = str(task.task_type_id)

                for entry in [entry_id, "total"]:
                    _add_quota_entry(
                        quotas,
                        entry,
                        date,
                        timezone,
                        nb_frames,
                        nb_drawings,
                        fps,
                    )
            date = date + timedelta(1)
    return quotas


def get_raw_quotas(
    project_id,
    task_type_id=None,
    person_id=None,
    studio_id=None,
    feedback=True,
):
    """
    Build quota statistics in a raw way. It counts the number of frames done
    for each day. A shot is considered done at the first feedback request (end
    date) or approval date (done_date).

    It considers that all the work was done at the end date.
    It computes the shot count and the number of seconds too.
    """
    fps = projects_service.get_project_fps(project_id)
    timezone = user_service.get_timezone()
    shot_type = get_shot_type()
    quotas = {}
    query = (
        Task.query.filter(Task.project_id == project_id)
        .filter(Entity.entity_type_id == shot_type["id"])
        .join(Entity, Entity.id == Task.entity_id)
        .join(Project, Project.id == Task.project_id)
        .join(Task.assignees)
        .add_columns(Entity.nb_frames, Person.id)
    )

    if task_type_id is not None:
        query = query.filter(Task.task_type_id == task_type_id)

    if person_id is not None:
        person = persons_service.get_person_raw(person_id)
        query = query.filter(Task.assignees.contains(person))

    if feedback:
        query = query.filter(Task.end_date != None)
    else:
        query = query.filter(Task.done_date != None)

    if studio_id is not None:
        persons_from_studio = Person.query.filter(
            Person.studio_id == studio_id
        ).all()
        query = query.filter(
            or_(*[Task.assignees.contains(p) for p in persons_from_studio])
        )

    result = query.all()

    for task, nb_frames, task_person_id in result:
        date = task.done_date
        if feedback:
            date = task.end_date

        if nb_frames is None:
            nb_frames = 0

        nb_drawings = task.nb_drawings or 0

        entry_id = str(task_person_id)
        if person_id is not None:
            entry_id = str(task.task_type_id)

        for entry in [entry_id, "total"]:
            _add_quota_entry(
                quotas, entry, date, timezone, nb_frames, nb_drawings, fps
            )
    return quotas


def _add_quota_entry(
    quotas, entry_id, date, timezone, nb_frames, nb_drawings, fps
):
    nb_seconds = nb_frames / fps
    date_str = date_helpers.get_simple_string_with_timezone_from_date(
        date, timezone
    )
    year = date_str[:4]
    week = year + "-" + str(date.isocalendar()[1])
    month = date_str[:7]
    if entry_id not in quotas:
        _init_quota_entry(quotas, entry_id)
    _init_quota_date(quotas, entry_id, date_str, week, month)
    quotas[entry_id]["day"]["frames"][date_str] += nb_frames
    quotas[entry_id]["day"]["seconds"][date_str] += nb_seconds
    quotas[entry_id]["day"]["drawings"][date_str] += nb_drawings
    quotas[entry_id]["day"]["count"][date_str] += 1
    quotas[entry_id]["week"]["frames"][week] += nb_frames
    quotas[entry_id]["week"]["seconds"][week] += nb_seconds
    quotas[entry_id]["week"]["drawings"][week] += nb_drawings
    quotas[entry_id]["week"]["count"][week] += 1
    quotas[entry_id]["month"]["frames"][month] += nb_frames
    quotas[entry_id]["month"]["seconds"][month] += nb_seconds
    quotas[entry_id]["month"]["drawings"][month] += nb_drawings
    quotas[entry_id]["month"]["count"][month] += 1
    quotas[entry_id]["year"]["frames"][year] += nb_frames
    quotas[entry_id]["year"]["drawings"][year] += nb_drawings
    quotas[entry_id]["year"]["seconds"][year] += nb_seconds
    quotas[entry_id]["year"]["count"][year] += 1


def _init_quota_date(quotas, entry_id, date_str, week, month):
    year = week[:4]
    if date_str not in quotas[entry_id]["day"]["frames"]:
        quotas[entry_id]["day"]["frames"][date_str] = 0
        quotas[entry_id]["day"]["seconds"][date_str] = 0
        quotas[entry_id]["day"]["count"][date_str] = 0
        quotas[entry_id]["day"]["drawings"][date_str] = 0
        if month not in quotas[entry_id]["day"]["entries"]:
            quotas[entry_id]["day"]["entries"][month] = 0
        quotas[entry_id]["day"]["entries"][month] += 1
    if week not in quotas[entry_id]["week"]["frames"]:
        quotas[entry_id]["week"]["frames"][week] = 0
        quotas[entry_id]["week"]["seconds"][week] = 0
        quotas[entry_id]["week"]["count"][week] = 0
        quotas[entry_id]["week"]["drawings"][week] = 0
        if year not in quotas[entry_id]["week"]["entries"]:
            quotas[entry_id]["week"]["entries"][year] = 0
        quotas[entry_id]["week"]["entries"][year] += 1
    if month not in quotas[entry_id]["month"]["frames"]:
        quotas[entry_id]["month"]["frames"][month] = 0
        quotas[entry_id]["month"]["seconds"][month] = 0
        quotas[entry_id]["month"]["count"][month] = 0
        quotas[entry_id]["month"]["drawings"][month] = 0
        if year not in quotas[entry_id]["month"]["entries"]:
            quotas[entry_id]["month"]["entries"][year] = 0
        quotas[entry_id]["month"]["entries"][year] += 1
    if year not in quotas[entry_id]["year"]["frames"]:
        quotas[entry_id]["year"]["frames"][year] = 0
        quotas[entry_id]["year"]["seconds"][year] = 0
        quotas[entry_id]["year"]["count"][year] = 0
        quotas[entry_id]["year"]["drawings"][year] = 0


def _init_quota_entry(quotas, entry_id):
    quotas[entry_id] = {
        "day": {
            "frames": {},
            "seconds": {},
            "count": {},
            "entries": {},
            "drawings": {},
        },
        "week": {
            "frames": {},
            "seconds": {},
            "count": {},
            "entries": {},
            "drawings": {},
        },
        "month": {
            "frames": {},
            "seconds": {},
            "count": {},
            "entries": {},
            "drawings": {},
        },
        "year": {"frames": {}, "seconds": {}, "count": {}, "drawings": {}},
    }


def get_month_quota_shots(
    person_id,
    year,
    month,
    project_id=None,
    task_type_id=None,
    weighted=True,
    feedback=True,
):
    """
    Return shots that are included in quota computation for given
    person and month.
    """
    start, end = date_helpers.get_month_interval(year, month)
    start, end = _get_timezoned_interval(start, end)
    if weighted:
        return get_weighted_quota_shots_between(
            person_id,
            start,
            end,
            project_id=project_id,
            task_type_id=task_type_id,
            feedback=feedback,
        )
    else:
        return get_raw_quota_shots_between(
            person_id,
            start,
            end,
            project_id=project_id,
            task_type_id=task_type_id,
            feedback=feedback,
        )


def get_week_quota_shots(
    person_id,
    year,
    week,
    project_id=None,
    task_type_id=None,
    weighted=True,
    feedback=True,
):
    """
    Return shots that are included in quota comptutation for given
    person and week.
    """
    start, end = date_helpers.get_week_interval(year, week)
    start, end = _get_timezoned_interval(start, end)
    if weighted:
        return get_weighted_quota_shots_between(
            person_id,
            start,
            end,
            project_id=project_id,
            task_type_id=task_type_id,
            feedback=feedback,
        )
    else:
        return get_raw_quota_shots_between(
            person_id,
            start,
            end,
            project_id=project_id,
            task_type_id=task_type_id,
            feedback=feedback,
        )


def get_day_quota_shots(
    person_id,
    year,
    month,
    day,
    project_id=None,
    task_type_id=None,
    weighted=True,
    feedback=True,
):
    """
    Return shots that are included in quota comptutation for given
    person and day.
    """
    start, end = date_helpers.get_day_interval(year, month, day)
    if weighted:
        return get_weighted_quota_shots_between(
            person_id,
            start,
            end,
            project_id=project_id,
            task_type_id=task_type_id,
            feedback=feedback,
        )
    else:
        return get_raw_quota_shots_between(
            person_id,
            start,
            end,
            project_id=project_id,
            task_type_id=task_type_id,
            feedback=feedback,
        )


def get_weighted_quota_shots_between(
    person_id, start, end, project_id=None, task_type_id=None, feedback=True
):
    """
    Get all shots leading to a quota computation during the given period.
    Set a weight on each one:
        * If there is time spent filled, weight it by the sum of duration
          divided py the overall task duration.
        * If there is no time spent, weight it by the number of business days
          in the time interval spent between WIP date (start) and
          feedback date (end).
    """
    shot_type = get_shot_type()
    person = persons_service.get_person_raw(person_id)
    shots = []
    already_listed = {}

    query = (
        Entity.query.filter(Entity.entity_type_id == shot_type["id"])
        .filter(Task.project_id == project_id)
        .filter(Task.task_type_id == task_type_id)
        .filter(TimeSpent.person_id == person_id)
        .filter(TimeSpent.date >= func.cast(start, TimeSpent.date.type))
        .filter(TimeSpent.date < func.cast(end, TimeSpent.date.type))
        .join(Task, Entity.id == Task.entity_id)
        .join(Project, Project.id == Task.project_id)
        .join(TimeSpent, Task.id == TimeSpent.task_id)
        .add_columns(Task.duration, TimeSpent.duration)
    )

    if feedback:
        query = query.filter(Task.end_date != None)
    else:
        query = query.filter(Task.done_date != None)

    query_shots = query.all()
    for entity, task_duration, duration in query_shots:
        shot = entity.serialize()
        if shot["id"] not in already_listed:
            full_name, _, _ = names_service.get_full_entity_name(shot["id"])
            shot["full_name"] = full_name
            shot["weight"] = round(duration / task_duration, 2) or 0
            shots.append(shot)
            already_listed[shot["id"]] = shot
        else:
            shot = already_listed[shot["id"]]
            shot["weight"] += round(duration / task_duration, 2)

    if type(start) is str:
        start = date_helpers.get_datetime_from_string(start)
    if type(end) is str:
        end = date_helpers.get_datetime_from_string(end)
    query = (
        Entity.query.filter(Entity.entity_type_id == shot_type["id"])
        .filter(Task.project_id == project_id)
        .filter(Task.task_type_id == task_type_id)
        .filter(Task.real_start_date != None)
        .filter(Task.assignees.contains(person))
        .filter(TimeSpent.id == None)
        .join(Task, Entity.id == Task.entity_id)
        .join(Project, Project.id == Task.project_id)
        .outerjoin(TimeSpent, TimeSpent.task_id == Task.id)
    )

    if feedback:
        query = (
            query.filter(Task.end_date != None)
            .filter((Task.real_start_date <= end) & (Task.end_date >= start))
            .add_columns(Task.real_start_date, Task.end_date)
        )
    else:
        query = (
            query.filter(Task.done_date != None)
            .filter((Task.real_start_date <= end) & (Task.done_date >= start))
            .add_columns(Task.real_start_date, Task.done_date)
        )

    query_shots = query.all()

    for entity, task_start, task_end in query_shots:
        shot = entity.serialize()
        if shot["id"] not in already_listed:
            business_days = (
                date_helpers.get_business_days(task_start, task_end) + 1
            )
            full_name, _, _ = names_service.get_full_entity_name(shot["id"])
            shot["full_name"] = full_name
            multiplicator = 1
            if task_start >= start and task_end <= end:
                multiplicator = business_days
            elif task_start >= start:
                multiplicator = (
                    date_helpers.get_business_days(task_start, end) + 1
                )
            elif task_end <= end:
                multiplicator = (
                    date_helpers.get_business_days(start, task_end) + 1
                )
            shot["weight"] = round(multiplicator / business_days, 2)
            already_listed[shot["id"]] = True
            shots.append(shot)

    return sorted(shots, key=itemgetter("full_name"))


def get_raw_quota_shots_between(
    person_id, start, end, project_id=None, task_type_id=None, feedback=True
):
    """
    Get all shots leading to a quota computation during the given period.
    """
    shot_type = get_shot_type()
    person = persons_service.get_person_raw(person_id)
    shots = []

    query = (
        Entity.query.filter(Entity.entity_type_id == shot_type["id"])
        .filter(Task.project_id == project_id)
        .filter(Task.task_type_id == task_type_id)
        .filter(Task.assignees.contains(person))
        .join(Task, Entity.id == Task.entity_id)
        .join(Project, Project.id == Task.project_id)
    )

    if feedback:
        query = query.filter(
            Task.end_date.between(
                func.cast(start, Task.end_date.type),
                func.cast(end, Task.end_date.type),
            )
        )
    else:
        query = query.filter(
            Task.done_date.between(
                func.cast(start, Task.done_date.type),
                func.cast(end, Task.done_date.type),
            )
        )

    query_shots = query.all()

    for entity in query_shots:
        shot = entity.serialize()
        full_name, _, _ = names_service.get_full_entity_name(shot["id"])
        shot["full_name"] = full_name
        shot["weight"] = 1
        shots.append(shot)

    return sorted(shots, key=itemgetter("full_name"))


def _get_timezoned_interval(start, end):
    """
    Get time intervals adapted to the user timezone.
    """
    timezone = user_service.get_timezone()
    return date_helpers.get_timezoned_interval(start, end, timezone)


def get_all_raw_shots():
    """
    Get all assets from the database.
    """
    query = Entity.query.filter(Entity.entity_type_id == get_shot_type()["id"])
    return query.all()


def set_frames_from_task_type_preview_files(
    project_id,
    task_type_id,
    episode_id=None,
):
    from zou.app import db

    shot_type = get_shot_type()
    Shot = aliased(Entity)
    Sequence = aliased(Entity)

    if episode_id is not None:
        subquery = (
            db.session.query(
                Shot.id.label("entity_id"),
                func.max(PreviewFile.created_at).label("max_created_at"),
            )
            .join(Task, PreviewFile.task_id == Task.id)
            .join(Shot, Task.entity_id == Shot.id)
            .join(Sequence, Sequence.id == Shot.parent_id)
            .filter(Shot.project_id == project_id)
            .filter(Shot.entity_type_id == shot_type["id"])
            .filter(Task.task_type_id == task_type_id)
            .filter(Sequence.parent_id == episode_id)
            .group_by(Shot.id)
            .subquery()
        )
    else:
        subquery = (
            db.session.query(
                Shot.id.label("entity_id"),
                func.max(PreviewFile.created_at).label("max_created_at"),
            )
            .join(Task, PreviewFile.task_id == Task.id)
            .join(Shot, Task.entity_id == Shot.id)
            .filter(Shot.project_id == project_id)
            .filter(Shot.entity_type_id == shot_type["id"])
            .filter(Task.task_type_id == task_type_id)
            .group_by(Shot.id)
            .subquery()
        )

    query = (
        db.session.query(Shot, PreviewFile.duration)
        .join(Task, Task.entity_id == Shot.id)
        .join(subquery, (Shot.id == subquery.c.entity_id))
        .join(
            PreviewFile,
            (PreviewFile.task_id == Task.id)
            & (PreviewFile.created_at == subquery.c.max_created_at),
        )
        .filter(Task.task_type_id == task_type_id)
        .filter(Shot.project_id == project_id)
    )

    results = query.all()
    project = projects_service.get_project(project_id)
    updates = []
    for shot, preview_duration in results:
        nb_frames = round(preview_duration * float(project["fps"]))
        updates.append(
            {
                "id": shot.id,
                "nb_frames": nb_frames,
            }
        )
        clear_shot_cache(str(shot.id))

    db.session.bulk_update_mappings(Shot, updates)
    db.session.commit()
    return updates



"""add prefeence fields

Revision ID: 8e67c183bed7
Revises: 59a7445a966c
Create Date: 2024-09-26 10:48:45.678791

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "8e67c183bed7"
down_revision = "59a7445a966c"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("organisation", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("dark_theme_by_default", sa.Boolean(), nullable=True)
        )
        batch_op.add_column(
            sa.Column("format_duration_in_hours", sa.Boolean(), nullable=True)
        )

    with op.batch_alter_table("project", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "is_publish_default_for_artists", sa.Boolean(), nullable=True
            )
        )
        batch_op.add_column(
            sa.Column("hd_bitrate_compression", sa.Integer(), nullable=True)
        )
        batch_op.add_column(
            sa.Column("ld_bitrate_compression", sa.Integer(), nullable=True)
        )

    with op.batch_alter_table("task", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "difficulty", sa.Integer(), nullable=False, server_default="3"
            )
        )
        batch_op.create_check_constraint(
            "check_difficulty", "difficulty > 0 AND difficulty < 6"
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("task", schema=None) as batch_op:
        batch_op.drop_constraint("check_difficulty", type_="check")
        batch_op.drop_column("difficulty")

    with op.batch_alter_table("project", schema=None) as batch_op:
        batch_op.drop_column("ld_bitrate_compression")
        batch_op.drop_column("hd_bitrate_compression")
        batch_op.drop_column("is_publish_default_for_artists")

    with op.batch_alter_table("organisation", schema=None) as batch_op:
        batch_op.drop_column("format_duration_in_hours")
        batch_op.drop_column("dark_theme_by_default")

    # ### end Alembic commands ###
